"""
Miscellaneous functions

"""

import numpy as np

#=================================
# Basic geochronology functions
#=================================

# def dc2hl():
#     """Convert decay constant to half life.
#     """
#     return
#
#
# def hl2dc():
#     """ Convert half life to decay constant.
#     """
#     return


#========
# Maths
#========

def pos_def(x, tol=1e-8):
    """
    Verify that matrix is positive semi-definite.
    Based on numpy multivariate_normal code, e.g.:
    https://github.com/numpy/numpy/blob/main/numpy/random/mtrand.pyx
    """
    x = np.asarray(x, dtype=np.double)
    (u, s, v) = np.linalg.svd(x)
    psd = np.allclose(np.dot(v.T * s, v), x, rtol=tol, atol=tol)
    return psd


def cdiff(x, f, h, *args, **kwargs):
    """
    Estimate derivative using central difference method
    """
    return (f(x + h, *args, **kwargs) - f(x - h, *args, **kwargs)) / (2.0 * h)


def eigsorted(cov):
    """
    Return eigenvalues and vectors sorted from largest to smallest.
    Based on code by Joe Kington:
    https://github.com/joferkington/oost_paper_code/blob/master/error_ellipse.py
    """
    vals, vecs = np.linalg.eigh(cov)
    order = vals.argsort()[::-1]
    return vals[order], vecs[:, order]


def covmat_to_cormat(cov):
    """
    Compute correlation matrix from covariance matrix. Based on:
    https://gist.github.com/wiso/ce2a9919ded228838703c1c7c7dad13b
    """
    v = np.sqrt(np.diag(cov))
    outer_v = np.outer(v, v)
    cor = cov / outer_v
    cor[cov == 0] = 0
    return cor


#=========================
# Number formatting etc.
#=========================

def get_exponent(n):
    """Return order of magnitude of n
    """
    if n == 0:
        return 0. * n
    return int(np.floor(np.log10(abs(n))))


def num_dec_places(n):
    """Return number of decimal digits in string representation of a decimal
    number.
    """
    s = str(n).rstrip("0")
    try:
        nd = s.split('.')[1]
    except IndexError:
        return 0
    return len(nd)


def round_down(n, sf=2):
    """ Round float down to n significant figures.
    """
    if n == 0:
        return 0. * n   # for array inputs
    return round(n, sf - int(np.floor(np.log10(abs(n)))) - 1)


#=========================
# Very basic I/O
#=========================

def print_result(dic, title='Result'):
    print(title)
    print('--------------------')
    for k, v in dic.items():
        print(f'{k}: {v}')
    print()


def print_table(cols, title='Table'):
    n = len(cols[0])
    assert [len(x) == n for x in cols]
    print(title)
    print('--------------------')
    for row in range(n):
        out = []
        for x in cols:
            out.append(f'{x[row]}')
        print(out)


"""
Disequlibrium Pb/U functions based on Wendt and Carl (1985)

References
----------
.. [Wendt1985]
    Wendt, I., Carl, C., 1985. U/Pb dating of discordant 0.1 Ma old secondary U
    minerals. Earth and Planetary Science Letters 73, 278–284.

"""

import numpy as np

from .. import cfg
from .. import useries


def f(t, A, DC, init=(True, True)):
    """
    206Pb/236U atomic ratio as a function of time and actvitiy
    ratios using equation from Wendt and Carl (1985)
    """
    assert cfg.secular_eq

    # Get initial activity ratios if present value given.
    A48i = A[0] if init[0] else useries.ar48i(t, A[0], *DC[:2])
    A08i = A[1] if init[1] else useries.ar08i(t, A48i, A[1], *DC[:3], init=True)

    A0 = A48i - 1
    B0 = A08i - 1
    C0 = A[2] - 1

    K1 = -A0 * DC[0] / DC[1] * ((DC[2] * DC[3]) * (DC[2] - DC[1]) * (DC[3] - DC[1]))
    K2 = DC[0] * DC[3] / (DC[3] - DC[2]) * (A0 / (DC[2] - DC[1]) - B0 / DC[2])
    K3 = DC[0] / (DC[3] - DC[2]) * (B0 - A0 * DC[2] / (DC[3] - DC[1])) - C0 * (DC[0] / DC[3])
    K4 = A0 * (DC[0] / DC[1]) + B0 * (DC[0] / DC[2]) + C0 * (DC[0] / DC[3])

    f2 = (np.exp(DC[0] * t) - 1) + np.exp(DC[0] * t) * (K1 * np.exp(-DC[1] * t)
            + K2 * np.exp(-DC[2] * t) + K3 * np.exp(-DC[3] * t) + K4)

    return f2


def g(t, A15i, DC):
    """
    207Pb/235U atomic ratio as a function of time and actvitiy
    ratios using equation from Wendt and Carl (1985).
    """
    D0 = A15i - 1.
    f1 = np.exp(DC[0] * t) - 1 + D0 * (DC[0] / DC[1]) * np.exp(DC[0] * t) \
         * (1 - np.exp(-DC[1] * t))
    return f1



"""
Functions and routines for calculating U-Pb ages **corrected** for
initial disequilibrium under the assumption that radioactive equilibrium
has been established at the time of analysis.

!!! This module contains experimental functions and routines that haven't
been properly tested or debugged !!!

Notes
-------
See dqpb module for disequilibrium U-Pb age functions - to be used if the sample
cannot be assumed to have attained radioactive equilbrium at the time
of analysis.

References
----------
.. [Mclean2011]
    McLean, N.M., Bowring, J.F., Bowring, S.A., 2011. An algorithm for U-Pb
    isotope dilution data reduction and uncertainty propagation. Geochemistry,
    Geophysics, Geosystems 12. https://doi.org/10.1029/2010GC003478
.. [Scharer1984]
    Schärer, U., 1984. The effect of initial $^{230}$Th disequilibrium on young
    U-Pb ages: the Makalu case, Himalaya. Earth and Planetary Science Letters
    67, 191–204.

"""

import numpy as np

from .. import cfg
from .. import plotting
from .. import wtd_average
from .. import monte_carlo as mc
from .. import misc


#=================================
# Single analysis Pb/U ages
#=================================

def pb206_age(x, sx, ThU_type=1, DThU=None, DThU_se=None, ThU_min=None,
        ThU_melt=None, ThU_melt_se=None, wav=False, dc_errors=False,
        sx_external=0., cov=True, wav_method='ca'):

    """
    !!! Experimental function !!!
    Compute 206Pb/238U ages corrected for initial 230Th/238U disequilbrium,
    under the assumption that the 238U-206Pb decay series has established
    radioactive equilbrium at the time of analysis (e.g. Schärer, 1984).
    """
    assert ThU_type in (1, 2)
    lam238, lam230 = cfg.lam238, cfg.lam230
    s238, s230 = cfg.s238, cfg.s230
    cov80 = cfg.cor_238_230 * s238 * s230
    n = x.shape[0]

    # rename parameters
    D = DThU
    sD = DThU_se
    rmin = ThU_min
    rmelt = ThU_melt
    srmelt = ThU_melt_se

    if ThU_type == 1:
        f = np.broadcast_to(D, (n,))          # note, f = f_ThU = (Th/U_min) / (Th/U_mag)
    else:
        f = rmin / rmelt

    t = -1. / lam238 * np.log(1. - x + lam238 / lam230 * (f - 1.))

    # compute age uncertainty
    dtx = np.exp(lam238 * t) / lam238            # dt/dx etc.

    if dc_errors:
        dt8 = - 1. / lam238 * (t + (f - 1) * np.exp(lam238 * t) / lam230)
        dt0 = (f - 1.) * np.exp(lam238 * t) / lam230 ** 2
    else:
        dt8 = np.zeros(n)
        dt0 = np.zeros(n)

    # Compile age covariance matrix following, e.g., McLean et al., (2011).
    # random errors:
    cov_tr = np.diag((dtx * sx) ** 2)

    # systematic errors:
    cov_s = np.zeros((4, 4))
    cov_s[-2:, -2:] = np.array([[s238 ** 2, cov80], [cov80, s230 ** 2]])
    cov_s[0, 0] = sx_external ** 2

    if ThU_type == 1:
        cov_s[1, 1] = sD ** 2
        dtD = - np.exp(lam238 * t) / lam230
        jac = np.stack((dtx, dtD, dt8, dt0), axis=-1)
    else:
        cov_s[1, 1] = srmelt ** 2
        dtrm = rmin * np.exp(lam238 * t) / (lam230 * rmelt ** 2)     # dt / d(rmelt)
        jac = np.stack((dtx, dtrm, dt8, dt0), axis=-1)

    cov_ts = jac @ cov_s @ jac.T

    # combined errors:
    cov_t = cov_tr + cov_ts
    st = np.sqrt(np.diag(cov_t))

    results = {
        'age': t,
        'age_1s': st,
        'age_95ci': 1.96 * st,
        'cov_t': cov_t,
        'cor_t': misc.covmat_to_cormat(cov_t)
    }

    if wav:
        if not cov:
            cov_t = np.diag(np.diag(cov_t))
        if wav_method in ('ca'):
            wav_results = wtd_average.classical_wav(t, cov=cov_t,
                                method=wav_method)
        elif wav_method in ('ra', 'sp', 'r2'):
            wav_results = wtd_average.robust_wav(t, cov=cov_t, method=wav_method)
        else:
            raise ValueError('wav method not recognised')

        results['wav_age'] = wav_results['ave']
        results['wav_age_95ci'] = wav_results['ave_95ci']
        results['wav'] = wav_results

        # Make wav plot:
        fig = plotting.wav_plot(t, results['age_95ci'], wav_results['ave'],
                    wav_results['ave_95ci'], sorted=sorted)
        ax = fig.get_axes()[0]
        ax.set_ylabel('$^{206}$Pb/$^{238}$U age (Ma)')
        results['wav_plot'] = fig

    return results


def pb207_age():
    return


#==========================
# Concordia-intercept age
#==========================

def concint_age():
    return

def isochron_age():
    return


#=================================
# Monte Carlo age uncertainties
#=================================

def mc_pb206_cor(t, x, sx, ThU_type=1, DThU=None, DThU_se=None, ThU_min=None,
            ThU_melt=None, ThU_melt_se=None, trials=10_000,  dc_errors=False,
            negative_ages=False, rand=False, hist=(False, False),
            wav_method='ca', wav=False):
    """
    !!! Experimental function !!!
    Compute age uncertainties for corrected Pb/U age using Monte Carlo
    approach.
    """
    assert not any([np.isnan(t) for t in t])       # nan ages should be removed

    n = x.size
    flags = np.zeros((trials, n), dtype='uint8')
    ts = np.empty((trials, n))                  # pre-allocate ages array

    # simulate fThU values
    if ThU_type == 1:
        fThU = cfg.rng.normal(DThU, DThU_se, trials)
        fThU = np.broadcast_to(fThU.reshape(trials, 1), (trials, n))
    elif ThU_type == 2:
        ThU_melt = cfg.rng.normal(ThU_melt, ThU_melt_se, (trials, 1))
        fThU = np.broadcast_to(ThU_min, (trials, n)) / ThU_melt

    if dc_errors:
        lam238 = cfg.rng.normal(cfg.lam238, cfg.s238, trials)
        lam230 = cfg.rng.normal(cfg.lam230, cfg.s230, trials)
    else:
        lam238 = cfg.lam238
        lam230 = cfg.lam230

    for i in range(n):
        xs = cfg.rng.normal(x[i], sx[i], trials)
        ts[:, i] = -1. / lam238 * np.log(1. - xs + lam238 / lam230 * (fThU[:, i] - 1.))

    good = np.all(flags == 0, axis=1)
    age_95ci = [np.quantile(t[good], (0.025, 0.975)) for t in ts.T]

    # compile results
    results = {
        'age_type': 'corrected 206Pb',
        'age_1s': [np.std(t[good]) for t in ts.T],
        'age_pm_95ci': [np.mean((t - m, p - t)) for m, p in age_95ci],
        'age_95ci': [np.quantile(t[good], (0.025, 0.975)) for t in ts.T],
        'mean_age': [np.mean(t[good]) for t in ts.T],
        'median_age': [np.median(t[good]) for t in ts.T],
        'cov_t': np.cov(np.transpose(ts[good])),
        'cor_t': np.corrcoef(np.transpose(ts[good])),
        'trials': trials,
        'fails': sum(flags != 0),
        'not_converged': sum(flags == mc.NON_CONVERGENCE),
        'negative_ages': sum(flags == mc.NEGATIVE_AGE)
    }

    if wav:
        cov_t = results['cov_t']
        if wav_method in ('ca'):
            wav_results = wtd_average.classical_wav(t, cov=cov_t,
                                method=wav_method)
        elif wav_method in ('ra', 'sp', 'r2'):
            wav_results = wtd_average.robust_wav(t, cov=cov_t, method=wav_method)
        else:
            raise ValueError('wav method not recognised')

        results['wav_age'] = wav_results['ave']
        results['wav_age_95ci'] = wav_results['ave_95ci']
        results['wav'] = wav_results

        # Make wav plot:
        fig = plotting.wav_plot(t, results['age_pm_95ci'], wav_results['ave'],
                    wav_results['ave_95ci'], sorted=sorted)
        ax = fig.get_axes()[0]
        ax.set_ylabel('$^{206}$Pb/$^{238}$U age (Ma)')
        results['wav_plot'] = fig

    return results



"""
Basic stats functions.

References
----------
.. [Ludwig2012]
    Ludwig, K.R., 2012. Isoplot/Ex Version 3.75: A Geochronological Toolkit for
    Microsoft Excel, Special Publication 4. Berkeley Geochronology Center.
.. [Wendt1991]
    Wendt, I., Carl, C., 1991. The statistical distribution of the mean squared
    weighted deviation. Chemical Geology: Isotope Geoscience section 86,
    275–285. https://doi.org/10.1016/0168-9622(91)90010-T

"""

import numpy as np
from scipy import stats


def nmad(x):
    """
    Median absolute deviation about the median (MAD) normalised such that
    statistic will be equal to the standard deviation for a standard normal
    random variable.
    """
    return 1.4826 * np.median(np.absolute(x - np.median(x)))


def t_critical(df, alpha=0.05):
    """
    Critical Student's t value for df degrees of freedom.
    """
    q = 1. - alpha / 2
    return stats.t.ppf(q, df)


def pr_fit(df, mswd):
    """
    Probability of fit following Ludwig (2012).
    """
    return 1. - stats.chi2.cdf(mswd * df, df)


def mswd_conf_limits(df, p=0.95, two_tailed=False):
    """
    Returns confidence limits on MSWD.
    """
    assert 0 < p < 1
    if two_tailed:
        p_low = (1. - p) / 2
        p_high = 1. - p_low
        lower = stats.distributions.chi2.ppf(p_low, df) / df
        upper = stats.distributions.chi2.ppf(p_high, df) / df
        return (lower, upper)
    else:
        # note: Wendt and Carl (1991) approximation is 1 + 2 * np.sqrt(2/df) ?
        upper = stats.distributions.chi2.ppf(p, df) / df
        return upper




"""
Functions and routines for U-Pb geochronology.

Notes
-----
See pysoplot.uxpb for equivalent functions that account for
disequilibrium in the uranium-series decay chains.

References
----------
.. [Horstwood2016]
    Horstwood, M.S.A., Košler, J., Gehrels, G., Jackson, S.E., McLean, N.M., Paton,
    C., Pearson, N.J., Sircombe, K., Sylvester, P., Vermeesch, P., Bowring, J.F.,
    Condon, D.J., Schoene, B., 2016. Community-derived standards for LA-ICP-MS
    U-(Th-)Pb geochronology - uncertainty propagation, age interpretation and
    data reporting. Geostandards and Geoanalytical Research 40, 311–332.
    https://doi.org/10.1111/j.1751-908X.2016.00379.x
.. [Ludwig1980]
    Ludwig, K.R., 1980. Calculation of uncertainties of U-Pb isotope data.
    Earth and Planetary Science Letters 212–202.
.. [Ludwig2000]
    Ludwig, K.R., 2000. Decay constant errors in U–Pb concordia-intercept ages.
    Chemical Geology 166, 315–318. https://doi.org/10.1016/S0009-2541(99)00219-3
.. [Powell2020]
    Powell, R., Green, E.C.R., Marillo Sialer, E., Woodhead, J., 2020. Robust
    isochron calculation. Geochronology 2, 325–342.
    https://doi.org/10.5194/gchron-2-325-2020

"""

import warnings
import numpy as np
import matplotlib.pyplot as plt

from scipy.optimize import newton

from . import isochron
from . import cfg
from . import transform
from . import monte_carlo as mc
from . import plotting
from . import exceptions

from .concordia import conc_xy, conc_slope, conc_age_x, plot_concordia


diagram_names = {
    'tw': 'Tera-Wasserburg',
    'wc': 'Wetheril Concordia'
}


def concint_age(fit, method='L1980', diagram='tw', dc_errors=False):
    """
    Compute concordia-intercept age and uncertainty using various algorithms.

    Parameters
    ----------
    fit : :obj:`dict`
        Regression fit parameters.
    method : {'Powell', 'L1980', 'L2000'}, optional
        Algorithm used to compute age and age uncertainties.

    Returns
    -------
    results : dict
        Age and age uncertainty results.
        `t_95ci` is uncertainty given as a 95% confidence interval.
        `t _95pm` is uncertainty given as a symmetric 95% confidence plus/minus error.

    Notes
    -----
    The 'Powell' method focusses on lower intercept age only (< ~100 Ma),
    but works with all fit type. Age uncertainties are symmetric. The 'L1980' method
    only works with classical fits. Age uncertainties are asymmetric. The 'L2000'
    method works with all fit types and optionally accounts for decay constant
    errors, but requires 2 intercept age solutions.

    """
    assert method in ('Powell', 'L1980', 'L2000')
    assert diagram in ('tw', 'wc')

    if fit['type'] == 'robust' and method == 'L1980':
        raise ValueError('cannot implement L1980 for data fitted with a '
                         'robust regression model')
    if dc_errors and method != 'L2000':
        raise ValueError('decay constant errors can only be included with '
                         'method L2000')

    # initialise vars for storing optional results:
    t = None
    t_95ci = None
    t2 = None
    t2_95pm = None
    t2_95ci = None
    cor_t12 = None

    if method == 'Powell':
        # Only compute lower intercept age
        t, st = concint_powell(fit['theta'], fit['covtheta'], uncert=True)
        t_95pm = fit['t_crit'] * st

    else:
        try:    # lower intercept
            t = concint_ludwig(fit['theta'], diagram=diagram, t0=-100.,
                               maxiter=40)
        except:
            warnings.warn('lower intercept age did not converge after max '
                          'number of iterations')
        try:    # upper intercept
            t2 = concint_ludwig(fit['theta'], diagram=diagram, t0=5000.,
                                maxiter=40)
        except:
            warnings.warn('upper intercept age did not converge after max '
                          'number of iterations')

        # check intercept ages
        if t is None:
            if t2 is not None:
                t = t2
                t2 = None
            else:
                raise RuntimeError('no intercept ages found')

        # compute age uncertainties
        if method == 'L1980':
            # lower intercept age:
            t_95ci = concint_uncert_ludwig1980(t, fit['theta'], fit['theta_95ci'],
                    fit['xbar'], diagram=diagram)
            t_95pm = np.mean([t_95ci[1] - t, t - t_95ci[0]])
            # upper intercept age:
            if t2 is not None:
                t2_95ci = concint_uncert_ludwig1980(t2, fit['theta'], fit['theta_95ci'],
                        fit['xbar'], diagram=diagram)
                t2_95pm = np.mean([t2_95ci[1] - t2, t2 - t2_95ci[0]])

        else:
            # L2000:
            assert diagram in ('tw', 'wc')
            if None in (t, t2):
                raise ValueError('age uncertainties could not be computed using '
                            'Ludwig2000 method because < 2 intercept age '
                            'solutions found')
            if diagram == 'tw':     # transform fit to wc
                theta, covtheta = transform.transform_fit(fit, transform_to='wc')
            else:
                theta, covtheta = fit['theta'], fit['covtheta']
            t_95pm, t2_95pm, cov_t12 = concint_uncert_ludwig2000(t, t2, theta,
                         covtheta, fit['t_crit'], dc_errors=dc_errors)
            # correlation between t1 and t2:
            cor_t12 = cov_t12 / (t_95pm * t2_95pm)

    results = {
        'age_type': 'concordia-intercept',
        'diagram': diagram_names[diagram],
        'decay_const_errors': dc_errors,
        'age': t,
        'age_95ci': t_95ci,
        'age_95pm': t_95pm,
        'upper_age': t2,
        'upper_age_95ci': t2_95ci,
        'upper_age_95pm': t2_95pm,
        'cor': cor_t12,
    }

    return results


def isochron_age(fit, age_type='iso-Pb6U8', dc_errors=False, norm_isotope='204Pb'):
    """
    Classical U-Pb isochron age and uncertainty.
    """
    assert age_type in ('iso-Pb6U8', 'iso-Pb7U5')
    b_95pm = fit['theta_95ci'][1]
    t, t_95pm = isochron.age(fit['theta'][1], sb=b_95pm, age_type=age_type,
                             dc_errors=dc_errors)
    results = {
        'age_type': f'{age_type} isochron age',
        'age': t,
        'age_95pm': t_95pm,
    }
    return results


def pbu_ages(dp, age_type='Pb6U8', cov=False, Pb76=None, wav=False):
    """
    Pb/U ages and uncertainty.
    !!! Not yet fully coded !!!
    """
    assert age_type in ('Pb6U8', 'Pb7U5', 'mod-207Pb')
    st = None

    if age_type in ('Pb6U8', 'Pb7U5'):
        x, sx = dp
        n = x.shape[0]
        t, st = pbu_age(x, sx, age_type=age_type, uncert=True)
    else:
        # use for initial guess at diseq. mod207 ages only !!!
        assert Pb76 is not None
        x, sx, y, sy, r_xy = dp
        n = x.shape[0]
        a = Pb76
        b = (y - a) / x
        t = np.empty(x.size)
        for i in range(n):
            t[i] = concint_powell((a, b[i]), uncert=False, maxiter=30, diagram='tw')

    # compute age uncertainties
    results = {
        'age_type': age_type,
        'age': t,
        'age_1s': st,
        'n': n,
    }
    if wav:
        raise ValueError('not yet coded')

    return results


#=====================================
# Concordia-intercept age algorithms
#=====================================

def concint_ludwig(theta, diagram='tw', t0=10., maxiter=40):
    """Compute concordia-intercept age using routine in the appendix of Ludwig
    (1980).
    """
    a, b = theta
    t1 = t0
    i = 0

    while i < maxiter:
        i += 1
        t = t1
        M = conc_slope(t, diagram)
        fx, fy = conc_xy(t, diagram)
        B = fy - M * fx
        x = (a - B) / (M - b)
        t1 = conc_age_x(x, diagram)
        if np.isclose(t, t1, rtol=1e-08, atol=1e-08):
            return t1

    raise exceptions.ConvergenceError('concordia-intercept age routine failed '
                      'to converge after maximum number of iterations')


def concint_uncert_ludwig1980(t, theta, theta_95ci, xbar, diagram='wc',
          maxiter=40, disp=True):
    """
    Compute concordia-intercept age uncertainties using the approach of Ludwig (1980).
    The algorithm can account for asymmetric age uncertainties, but not decay
    constant errors.

    Notes
    -----
    This algorithm requires x-bar to be specified so it is really only suitable
    for use with classical regression lines.

    """
    a, b = theta
    sa, sb = theta_95ci
    t_95ci = []

    for sign in (1, -1):
        t0 = t
        i = 0
        while i < maxiter:
            i += 1
            M = conc_slope(t0, diagram)
            fx, fy = conc_xy(t0, diagram)
            B = fy - M * fx
            D = 2 *((B - a) * (M - b) + xbar * sb ** 2)
            E = (M - b) ** 2 - sb ** 2
            V = (B - a) ** 2 - sa ** 2
            x = (-D + sign * np.sqrt(D ** 2 - 4 * E * V)) / (2 * E)
            t1 = conc_age_x(x, diagram)

            if np.isclose(t1, t0, rtol=1e-08, atol=1e-08):
                t_95ci.append(t1)
                break

            t0 = t1

        if i == maxiter:
            raise exceptions.ConvergenceError('Ludwig concordia-intercept '
                      'age error routine did not converge after maximum number '
                      'of iterations')

    t_95ci.sort()   # just in case
    return t_95ci


def concint_powell(theta, covtheta=None, uncert=False, maxiter=30,
                   diagram='tw', disp=True):
    """
    Compute Tera-Wasserburg concordia intercept age and uncertainty focussing
    on lower intercept only using the method of Powell et al. (2020). Converges
    rapidly, but will not work for intercept ages greater than ~100 Ma.
    Does not account for decay constant errors.

    """
    if diagram == 'wc':
        raise ValueError('Wetheril concordia version of Powell algorithm not '
                 'yet implemented')
    if uncert and covtheta is None:
        raise ValueError('covtheta cannot be None if uncert is True')

    lam238, lam235, U = cfg.lam238, cfg.lam235, cfg.U
    a, b = theta
    if b > 0:
        warnings.warn('Powell algorithm mat not converge for positive '
                     'Tera-Wasserburg slope')
    i = 0
    t = x = 1500.

    while i < maxiter:
        i += 1
        tc = t
        t = 1. / lam238 * np.log(1. + 1. / x)
        x = (1. / U * (np.exp(lam235 * t) - 1.) /
             (np.exp(lam238 * t) - 1.) - a) / b

        if np.isclose(t, tc, rtol=1e-09):

            if t < 0:
                warnings.warn(f'negative lower intercept age {t:.3f} Ma - '
                              f'check input data')
            if not uncert:
                return t

            den = b * np.exp(lam238 * t) * lam238 + \
                  (-np.exp(lam235 * t) * lam235 + np.exp(lam235 * t
                    + lam238 * t) * lam235 + np.exp(lam238 * t) * lam238
                   - np.exp(lam235 * t + lam238 * t) * lam238) / U
            jac = np.array([(np.exp(lam238 * t) - 1.) / den * (np.exp(lam238 * t)
                        - 1.), (np.exp(lam238 * t) - 1.) / den])
            st = np.sqrt(jac @ covtheta @ jac.T)

            return t, st

    raise exceptions.ConvergenceError('concordi-intercept age routine did not '
              'converge after maximum number of iterations')


def concint_uncert_ludwig2000(t1, t2, theta, covtheta, t_mult, dc_errors=True):
    """
    Compute lower and upper concordia intercept age and uncertainties in one
    go based on equations in Ludwig (2000). Can account for decay constant
    errors. Assumes age uncertainties are symmetric.

    Equations require data points to be in Wetheril concordia form. A
    transformation should first be performed if data points are in
    Tera-Wasserburg form.

    Notes
    -----
    Ludwig (2000) uses 'm' for regression line slope and 'b' for regression
    line y-int, whereas this function uses 'b' for regression line slope and
    'a' for y-intercept.

    """
    assert t2 > t1
    lam238, lam235, U = cfg.lam238, cfg.lam235, cfg.U
    a, b = theta
    covtheta = t_mult ** 2 * covtheta

    # Define some useful quantities
    delta = np.exp(lam235 * t1) - np.exp(lam235 * t2)
    m1 = (lam238 * np.exp(lam238 * t1) - b * lam235 * np.exp(lam235 * t1)) / delta
    m2 = -(lam238 * np.exp(lam238 * t2)
           - b * lam235 * np.exp(lam235 * t2)) / delta
    b1 = -(np.exp(lam235 * t2) - 1.) * m1
    b2 = -(delta + (np.exp(lam235 * t2) - 1.)) * m2

    # Need to solve Ax = B (where from Ludwig's notation Omega -> A and
    # Y -> B):

    # initialise A and B as empty arrays
    A = np.empty((3, 3))
    B = np.empty((3, 1))

    # Eq. (13)
    A[0, 0] = m1 ** 2; A[0, 1] = m2 ** 2; A[0, 2] = 2 * m1 * m2
    A[1, 0] = b1 ** 2; A[1, 1] = b2 ** 2; A[1, 2] = 2 * b1 * b2
    A[2, 0] = m1 * b1; A[2, 1] = m2 * b2; A[2, 2] = m1 * b2 + m2 * b1

    # Eq (14) - regression line error components only
    B[0, 0] = covtheta[1, 1]
    B[1, 0] = covtheta[0, 0]
    B[2, 0] = covtheta[0, 1]

    # If dc_errors, add decay constant error components:
    if dc_errors:
        s238 = 2. * cfg.s238
        s235 = 2. * cfg.s235
        m3 = b * (t2 * np.exp(lam235 * t2) - t1 * np.exp(lam235 * t1)) / delta
        m4 = (t1 * np.exp(lam238 * t1) - t2 * np.exp(lam238 * t2)) / delta
        b3 = - b * t2 * np.exp(lam235 * t2) - (np.exp(lam235 * t2) - 1.) * m3
        b4 = t2 * np.exp(lam238 * t2) - (np.exp(lam235 * t2) - 1.) * m4

        # Eq. (14)
        B[0, 0] += (m3 * s235) ** 2 + (m4 * s238) ** 2
        B[1, 0] += (b3 * s235) ** 2 + (b4 * s238) ** 2
        B[2, 0] += m3 * b3 * s235 ** 2 + m4 * b4 * s238 ** 2

    # Solve for x, Eq. (12)
    x = np.linalg.inv(A) @ B
    t_95pm, t2_95pm, cov = np.sqrt(x[0][0]), np.sqrt(x[1][0]), x[2][0]
    return t_95pm, t2_95pm, cov


#==================================================================
# Single analysis Pb/U ages
#==================================================================

def pbu_age(x, sx, age_type='Pb6U8', dc_errors=False, uncert=False):
    """
    Single analysis Pb/U age.

    Notes
    -----
    Uncertainties in a suite of single-analysis Pb/U ages should be
    propagated by quadratic addition after taking the weighted mean (e.g.,
    Horstwood, 2016).

    """
    assert age_type in ('Pb6U8', 'Pb7U5')
    dc = cfg.lam238 if age_type == 'Pb6U8' else cfg.lam235
    sdc = cfg.s238 if age_type == 'Pb6U8' else cfg.s235
    t = 1. / dc * np.log(x + 1.)
    if not uncert:
        return t
    # Propagate age uncertainty...
    dx = (1. / dc) * 1. / (x + 1.)                        # dt/dx
    ddc = t / dc if dc_errors else 0.                     # dt/ddc
    st = np.sqrt((dx * sx) ** 2 + (ddc * sdc) ** 2)
    return t, st


#=================================
# Age minimisation functions
#=================================

def concint_age_min(t0=1.0, diagram='tw', A48_init=True, A08_init=True):
    """
    Concordia-intercept age minimisation functions for Monte Carlo
    simulation.
    """
    assert diagram in ('tw', 'wc')
    h = t0 * np.sqrt(np.finfo(float).eps)

    if diagram == 'tw':
        def f(t, a, b, lam238, lam235, U):
            return (b + a * (np.exp(lam238 * t) - 1.) - (np.exp(lam235 * t) - 1.) / U)
        def fprime(t, a, b, lam238, lam235, U):
            return (a * lam238 * np.exp(lam238 * t) - lam235 / U * np.exp(lam235 * t))

    else:
        raise ValueError('not yet implemented')

    return f, fprime


#==================================================================
# Monte Carlo errors
#==================================================================

def mc_concint(t, fit, trials=50_000, diagram='tw', dc_errors=False,
        U_errors=False, intercept_plot=False, hist=False,
        xlim=(None, None), ylim=(None, None), env=False,
        age_ellipses=False, marker_max=None, marker_ages=(), auto_marker_ages=True,
        remove_overlaps=True, intercept_points=True, intercept_ellipse=False,
        negative_ages=True, age_prefix='Ma'):
    """
    Equilibrium discordia intercept age errors.

    """
    assert diagram in ('tw', 'wc')

    if env and not dc_errors:
        warnings.warn('cannot plot equilbrium concordia envelope if dc_errors set to False')
        env = False
    if age_ellipses and not dc_errors:
        warnings.warn('cannot plot equilibrium age ellipse markers if dc_errors set to False')
        age_ellipses = False

    flags = np.zeros(trials)

    # draw randomised slope-intercept values
    (a, b), flags = mc.draw_theta(fit, trials, flags)

    # draw randomised constants
    if dc_errors:
        lam238 = cfg.rng.normal(cfg.lam238, cfg.s238, trials)
        lam235 = cfg.rng.normal(cfg.lam235, cfg.s235, trials)
    else:
        lam238, lam235 = cfg.lam238, cfg.lam235

    U = cfg.rng.normal(cfg.U, cfg.sU, trials) if U_errors else cfg.U

    # solve for age using vectorised Newton
    fmin, dfmin = concint_age_min(diagram='tw', t0=t)
    ts, c, zd = newton(fmin, np.full(trials, t), fprime=dfmin,
            args=(a, b, lam238, lam235, U), tol=1e-09, rtol=0, disp=False,
            maxiter=50, full_output=True)

    flags = mc.check_ages(ts, c, flags, negative_ages=negative_ages)
    ok = flags == 0
    age_95ci = np.quantile(ts[ok], (0.025, 0.975))

    results = {
        'age_type': 'concordia-intercept',
        'diagram': diagram_names[diagram],
        'age_1s': np.nanstd(ts[ok]),
        'age_95ci': age_95ci,
        'age_95pm': np.nanmean([t - age_95ci[0], age_95ci[1] - t]),
        'mean_age': np.nanmean(ts[ok]),
        'median_age': np.nanmedian(ts[ok]),
        'trials': trials,
        'fails': sum(flags != 0),
        'not_converged': sum(flags == mc.NON_CONVERGENCE),
        'negative_ages': sum(flags == mc.NEGATIVE_AGE)
    }

    if hist:
        fig = mc.age_hist(ts, diagram, a=a, b=b)
        results['age_hist'] = fig

    # make intercept plot:
    if intercept_plot:
        fig, ax = plt.subplots(**cfg.fig_kw)

        # x, y intercep points
        lams238 = lam238 if not dc_errors else lam238[ok]
        lams238 = lam238 if not dc_errors else lam238[ok]
        U = U if not U_errors else U[ok]
        x = 1. / (np.exp(lam238 * ts) - 1.)
        y = (np.exp(lam235 * ts) - 1.) * x / U

        if intercept_points:
            ax.plot(x, y, label="mc marker", **cfg.conc_intercept_markers_kw)
        if intercept_ellipse:
            cov = np.cov(x, y)
            e = plotting.confidence_ellipse2(ax, np.mean(x), np.mean(y), cov,
                                             **cfg.conc_intercept_ellipse_kw)
            ax.add_patch(e)

        mc.intercept_plot_ax_limits(ax, fit['theta'][1], x, y, diagram='tw')
        plotting.apply_plot_settings(fig, plot_type='intercept', diagram=diagram,
                                     xlim=xlim, ylim=ylim)

        # add other plot elements
        plotting.plot_rfit(ax, fit)
        plot_concordia(ax, diagram, plot_markers=True, env=env, 
                       age_ellipses=age_ellipses, marker_max=marker_max,
                       marker_ages=marker_ages, auto_markers=auto_marker_ages,
                       remove_overlaps=remove_overlaps, age_prefix=age_prefix)
        results['fig'] = fig

    return results


def mc_pbu():
    """ Not yet coded.
    """
    pass



"""
Functions and routines for computing disequilibrium U-Pb ages

References
----------
.. [Engel2019]
    Engel, J., Woodhead, J., Hellstrom, J., Maas, R., Drysdale, R., Ford, D.,
    2019. Corrections for initial isotopic disequilibrium in the speleothem
    U-Pb dating method. Quaternary Geochronology 54, 101009.
    https://doi.org/10.1016/j.quageo.2019.101009
.. [Guillong2014]
    Guillong, M., von Quadt, A., Sakata, S., Peytcheva, I., Bachmann, O., 2014.
    LA-ICP-MS Pb-U dating of young zircons from the Kos–Nisyros volcanic centre,
    SE aegean arc. Journal of Analytical Atomic Spectrometry 29, 963–970.
    https://doi.org/10.1039/C4JA00009A
.. [Ludwig1977]
    Ludwig, K.R., 1977. Effect of initial radioactive-daughter disequilibrium
    on U-Pb isotope apparent ages of young minerals. Journal of Research of
    the US Geological Survey 5, 663–667.
.. [Sakata2017]
    Sakata, S., Hirakawa, S., Iwano, H., Danhara, T., Guillong, M., Hirata, T.,
    2017. A new approach for constraining the magnitude of initial
    disequilibrium in Quaternary zircons by coupled uranium and thorium decay
    series dating. Quaternary Geochronology 38, 1–12.
    https://doi.org/10.1016/j.quageo.2016.11.002

"""

import numpy as np
from scipy import optimize

from . import ludwig
from . import minimise
from . import dqmc

from .. import useries
from .. import wtd_average
from .. import cfg
from .. import plotting
from .. import exceptions


exp = np.exp
log = np.log
nan = np.nan


#===========================================
# Age calculation routines
#===========================================

def concint_age(fit, A, sA, init, t0, diagram='tw', dc_errors=False, trials=50_000,
        u_errors=False, negative_ar=True, negative_ages=True,
        intercept_plot=True, hist=(False, False), conc_kw=None,
        intercept_plot_kw=None, A48i_lim=(0., 20.), A08i_lim=(0., 10.),
        age_lim=(0., 20.)):
    """
    Compute disequilibrium U-Pb concordia-intercept age and age uncertainties using
    Monte Carlo simulation. Optionally, produce a plot of the concordia-intercept.
    
    Parameters
    ----------
    fit : dict
        linear regression fit parameters
    A : array-like
        one-dimensional array of activity ratio values arranged as follows
        - [234U/238U], [230Th/238U], [226Ra/238U], [231Pa/235U]
    sA : array-like
        one-dimensional array of activity ratio value uncertainties given 
        as 1 sigma absolute and arranged in the same order as A
    init : array-like
        two-element list of boolean values, the first is True if [234U/238U]
        is an initial value and False if a present-day value, the second is True
        if [230Th/238U] is an initial value and False if a present-day value
    t0 : float
        initial guess for numerical age solving (the equilibrium age is
        typically good enough)
    
    """
    assert type(init[0]) == bool and type(init[1] == bool)

    if intercept_plot_kw is None:
        intercept_plot_kw = {}
    if conc_kw is None:
        conc_kw = {}

    # Compute age:
    a, b = fit['theta']
    
    # If a present-day activity ratio given, check for all possible intercept age
    # solutions within given age limits.
    if any(init):
        r = concint_multiple(a, b, A, init, t0, age_lim=age_lim,  t_step=1e-5,
                             A48i_lim=A48i_lim, A08i_lim=A08i_lim)
        ages, A48i, A08i = r
        t = ages[0]
    else:
        t = concint(a, b, A, init, t0)
        if t < 0:
            raise RuntimeError(f'negative disequilibrium age solution: {t} Ma')

    # Get initial acitivity rario solutions:
    A48i, A08i = useries.Ai_solutions(t, A[:-1], init, (cfg.lam238,
                            cfg.lam234, cfg.lam230))

    # Compute age uncertainties:
    mc = dqmc.concint_age(t, fit, A, sA, init, trials=trials, dc_errors=dc_errors,
                  u_errors=u_errors, negative_ar=negative_ar,
                  negative_ages=negative_ages,hist=hist,
                  intercept_plot=intercept_plot, conc_kw=conc_kw,
                  intercept_plot_kw=intercept_plot_kw)

    results = {
        'age_type': 'concordia-intercept',
        'diagram': diagram,
        'age': t,
        'age_1s': mc['age_1s'],
        'age_95ci': mc['age_95ci'],
        'age_95pm': np.mean((t - mc['age_95ci'][0], mc['age_95ci'][1] - t)),
        '[234U/238U]i': A48i,
        '[230Th/238U]i': A08i,
        'mc': mc,
    }

    if not init[0]:
        results['[234U/238U]i_95ci'] = mc['[234U/238U]i_95ci']
        results['[234U/238U]i_95pm'] = np.mean((t - mc['[234U/238U]i_95ci'][0],
                                      mc['[234U/238U]i_95ci'][1] - t))
    if not init[1]:
        results['[230Th/238U]i_95ci'] = mc['[230Th/238U]i_95ci']
        results['[230Th/238U]i_95pm'] = np.mean((t - mc['[230Th/238U]i_95ci'][0],
                                        mc['[230Th/238U]i_95ci'][1] - t))

    return results


def isochron_age(fit, A, sA, t0, init=(True, True), age_type='iso-Pb6U8',
                 norm_isotope='204Pb', hist=(False, False),
                 trials=50_000, dc_errors=False, negative_ar=True,
                 negative_ages=True):
    """
    Compute disequilibrium 238U-206Pb or 235U-207Pb isochron age and age
    uncertainties using Monte Carlo methods.

    Parameters
    ----------
    fit : dict
        linear regression fit parameters
    A : array-like
        one-dimensional array of activity ratio values arranged as follows
        - [234U/238U], [230Th/238U], [226Ra/238U], [231Pa/235U]
    sA : array-like
        one-dimensional array of activity ratio value uncertainties given
        as 1 sigma absolute and arranged in the same order as A
    init : array-like
        two-element list of boolean values, the first is True if [234U/238U]
        is an initial value and False if a present-day value, the second is True
        if [230Th/238U] is an initial value and False if a present-day value
    t0 : float
        initial guess for numerical age solving (the equilibrium age is
        typically good enough)

    """
    assert age_type in ('iso-Pb6U8', 'iso-Pb7U5')

    if age_type == 'iso-Pb6U8':
        assert len(A) == 3
        assert len(sA) == 3
    else:
        assert type(A) in (float, int)
        assert type(sA) in (float, int)
    
    # Compute age:
    a, b = fit['theta']
    t = isochron(b, A, t0, age_type, init=init)
    if t < 0:
        raise RuntimeError(f'negative disequilibrium age solution: {t} Ma')

    # Back-calculate activity ratios:
    if age_type == 'iso-Pb6U8':
        DC = (cfg.lam238, cfg.lam234, cfg.lam230, cfg.lam226)
        A48i, A08i = useries.Ai_solutions(t, A, init, DC)
    else:
        A48i, A08i = (None, None)

    # Compute age uncertainties:
    mc = dqmc.isochron_age(t, fit, A, sA, init=init, trials=trials,
                        negative_ar=negative_ar, negative_ages=negative_ages,
                        hist=hist, age_type=age_type, dc_errors=dc_errors)

    results = {
        'age_type': f'{age_type} isochron',
        'norm_isotope': norm_isotope,
        'age': t,
        'age_1s': mc['age_1s'],
        'age_95ci': mc['age_95ci'],
        'age_95pm': mc['age_95pm'],
        '[234U/238U]i': A48i,
        '[230Th/238U]i': A08i,
        'mc': mc,
    }

    return results


def pbu_age(x, sx, t0, ThU_min=None, ThU_melt=None, ThU_melt_1s=None,
        DThU=None, DThU_1s=None, DPaU=None, DPaU_1s=None, DThU_const=True,
        age_type='Pb6U8', wav=False, trials=30_000, wav_method='sp',
        method='Ludwig', wav_plot_prefix='ka', sorted=False, rand=True,
        cov=True, hist=(False, False), negative_ages=True, negative_ar=True,
        ylim=(None, None), dp_labels=None):
    """
    Compute disequilibrium 206Pb/238U or 207Pb/235U ages, and optionally compute
    a weighted average age.

    Parameters
    ----------
    x : np.ndarray
        one-dimensional array of Pb/U ratios
    sx : np.ndarray
        one-dimensional array of Pb/U ratio errors (1s, abs.)
    t0 : np.ndarray
        initial guesses for disequilibrium age solutions
    wav : bool
        compute weighted average age
    wav_type : str
        wtd. average type
    method : str
        method of age calculation
    rand : bool
        plot random age uncerts on wtd. average plot
    cov : bool
        account for age covariance structure in wtd. average

    """
    assert age_type in ('Pb6U8', 'Pb7U5')
    assert method in ('Ludwig', 'Guillong')
    n = x.shape[0]
    assert sx.shape[0] == n

    if age_type == 'Pb6U8':
        if DThU_const:
            DThU = float(DThU)
            DThU_1s = float(DThU_1s)
        else:
            assert len(ThU_min) == n
            ThU_melt = float(ThU_melt)
            ThU_melt_1s = float(ThU_melt_1s)
    else:
        DPaU = float(DPaU)
        DPaU_1s = float(DPaU_1s)

    if np.atleast_1d(t0).size:
        t0 = np.broadcast_to(t0, (n,))

    # equilibrium activity ratio values
    A48_eq = cfg.A48_eq
    A68_eq = cfg.A08_eq

    if age_type == 'Pb6U8':
        if DThU_const:
            fXU = np.broadcast_to(DThU, (n,))
        else:
            XU_melt = ThU_melt
            XU_min = ThU_min
            fXU = XU_min / XU_melt
    else:
        # only type 1 values for Pa/U
        DPaU = DPaU
        fXU = np.broadcast_to(DPaU, (n,))

    # Comupte ages:
    t = np.empty(n)
    for i in range(n):
        if method == 'Guillong':
            t[i] = guillong(x[i], fXU[i], t0[i], age_type)
        else:
            if age_type == 'Pb6U8':
                t[i] = pbu(x[i], [A48_eq, fXU[i], A68_eq], t0[i], age_type)
            else:
                t[i] = pbu(x[i], fXU[i], t0[i], age_type)

    if np.all(np.isnan(t)):
        raise ValueError('all age solutions failed to converge')

    # Compute age uncertainties:
    mc = dqmc.pbu_age(
        t, x, sx, age_type=age_type, ThU_min=ThU_min,
        ThU_melt=ThU_melt, ThU_melt_1s=ThU_melt_1s, DThU=DThU,
        DThU_1s=DThU_1s, DPaU=DPaU, DPaU_1s=DPaU_1s,
        DThU_const=DThU_const, method=method, negative_ages=negative_ages,
        negative_ar=negative_ar, rand=rand, trials=trials, hist=hist,
    )

    results = {
        'age_type': age_type,
        'age': t,
        'age_1s': mc['age_1s'],
        'age_95ci': mc['age_95ci'],
        'age_95pm': mc['age_95pm'],
        'cov_t': mc['cov_t'],
        'mc': mc
    }

    rand_pm = None

    # compute wtd. average
    if wav:
        cov_t = mc['cov_t']
        if not cov:
            cov_t = np.diag(np.diag(cov_t))
        elif rand:
            results['age_rand_95ci'] = mc['age_rand_95ci']
            results['age_rand_95pm'] = mc['age_rand_95pm']
            rand_pm = np.asarray(results['age_rand_95pm'])
        if wav_method in ('ca'):
            wav_results = wtd_average.classical_wav(t, cov=cov_t, method=wav_method)
        elif wav_method in ('ra', 'rs'):
            wav_results = wtd_average.robust_wav(t, cov=cov_t, method=wav_method)
        else:
            raise ValueError('wav method not recognised')

        results['wav_age'] = wav_results['ave']
        results['wav_age_95pm'] = wav_results['ave_95pm']
        results['wav'] = wav_results

        # Make wav plot:
        age_mult = 1. if wav_plot_prefix == 'Ma' else 1000.
        fig = plotting.wav_plot(t, np.asarray(results['age_95pm']), wav_results['ave'],
                    wav_results['ave_95pm'], rand_pm=rand_pm,
                    sorted=sorted, ylim=ylim, x_multiplier=age_mult,
                    dp_labels=dp_labels)
        ax = fig.get_axes()[0]
        if age_type == 'Pb6U8':
            ax.set_ylabel(f'$^{{206}}$Pb/$^{{238}}$U age ({wav_plot_prefix})')
        else:
            ax.set_ylabel(f'$^{{207}}$Pb/$^{{235}}$U age ({wav_plot_prefix})')
        results['fig_wav'] = fig

    return results


def mod207_age(dp, Pb76, Pb76se, t0, ThU_min=None, ThU_melt=None, ThU_melt_1s=None,
            DThU=None, DThU_1s=None, DPaU=None, DPaU_1s=None, DThU_const=True,
            wav=False, wav_method='sp', method='Ludwig',
            sorted=False, rand=True, cov=True, negative_ages=True,
            negative_ar=True, ylim=(None, None), hist=(False, False),
            wav_plot_prefix='Ma', trials=50_000, dp_labels=None):
    """
    Compute modified 207Pb ages and optionally compute a weighted average age.

    Parameters
    ----------
    dp : np.ndarray
        5 x n array of data points
    t0 : array-like
        initial disequilibrium age guesses
    wav : bool
        compute weighted average age
    wav_method : str
        wtd. average method - one of 'ca', 'ra', 'sp', 'r2'
    method : str
        method of age calculation, one of 'Ludwig' or 'Sakata'
    rand : bool
        plot random age uncerts on wtd. average plot
    cov : bool
        account for age covariance structure in wtd. average

    """
    assert method in ('Ludwig', 'Sakata')
    x, sx, y, sy, r_xy = dp
    n = x.shape[0]
    assert sx.shape[0] == y.shape[0] == sy.shape[0] == r_xy.shape[0] == n
    if DThU_const:
        DThU = float(DThU)
        DThU_1s = float(DThU_1s)
    else:
        assert len(ThU_min) == n
        ThU_melt = float(ThU_melt)
        ThU_melt_1s = float(ThU_melt_1s)
    DPaU = float(DPaU)
    DPaU_1s = float(DPaU_1s)

    if np.atleast_1d(t0).size:
        t0 = np.broadcast_to(t0, (n,))

    # secular eq. values should be fine for most cases
    A48_eq = cfg.A48_eq
    A68_eq = cfg.A68_eq

    if DThU_const:
        fThU = np.broadcast_to(DThU, (n,))
    else:
        fThU = ThU_min / ThU_melt

    # only type 1 useful for Pa/U?
    fPaU = np.broadcast_to(DPaU, (n,))

    # Comupte ages:
    t = np.zeros(n)
    for i in range(n):
        if method == 'Ludwig':
            A = [A48_eq, fThU[i], A68_eq, fPaU[i]]
            init = (True, True)     # no present values for now
        # try:
        if method == 'Sakata':
            t[i] = sakata(x[i], y[i], fThU[i], fPaU[i], Pb76, t0[i])
        else:
            t[i] = mod207(x[i], y[i], A, Pb76, t0[i])
        # except:
        #     t[i] = nan

    if np.all(np.isnan(t)):
        raise ValueError('all age solutions failed to converge')

    # Compute age uncertainties:
    # dc_errors not yet implemented
    mc = dqmc.mod207_age(t, x, sx, y, sy, r_xy, Pb76, Pb76se, ThU_min=ThU_min,
            ThU_melt=ThU_melt, ThU_melt_1s=ThU_melt_1s, DThU=DThU,
            DThU_1s=DThU_1s, DPaU=DPaU, DPaU_1s=DPaU_1s, DThU_const=DThU_const,
            trials=trials, method=method, rand=rand, negative_ages=negative_ages,
            negative_ar=negative_ar, hist=hist)

    results = {
        'age_type': 'mod-207Pb',
        'age': t,
        'age_1s': mc['age_1s'],
        'age_95ci': mc['age_95ci'],
        'age_95pm': np.mean(abs(np.asarray(mc['age_95ci']) - t.reshape(n, -1)), axis=1),
        'cov_t': mc['cov_t'],
        'mc': mc
    }

    rand_pm = None

    # compute wtd. average
    if wav:
        if cov:
            cov_t = mc['cov_t']
            results['cov_t'] = cov_t
            wav_args = dict(cov=cov_t, method=wav_method)
        else:
            wav_args = dict(sx=np.diag(np.diag(mc['cov_t'])))
        if rand:
            results['age_rand_95ci'] = mc['age_rand_95ci']
            results['age_rand_95pm'] = np.mean(abs(np.asarray(mc['age_rand_95ci'])
                                                   - t.reshape(n, -1)), axis=1)
            rand_pm = results['age_rand_95pm']
        if wav_method in ('ca'):
            wav_results = wtd_average.classical_wav(t, **wav_args)
        elif wav_method in ('ra', 'rs', 'r2'):
            wav_results = wtd_average.robust_wav(t, **wav_args)
        else:
            raise ValueError('wav method not recognised')

        results['wav_age'] = wav_results['ave']
        results['wav_age_95pm'] = wav_results['ave_95pm']
        results['wav'] = wav_results

        # Make wav plot:
        age_mult = 1. if wav_plot_prefix == 'Ma' else 1000.
        fig = plotting.wav_plot(t, results['age_95pm'], wav_results['ave'],
                    wav_results['ave_95pm'], rand_pm=rand_pm,
                    sorted=sorted, ylim=ylim, x_multiplier=age_mult,
                    dp_labels=dp_labels)
        ax = fig.get_axes()[0]
        ax.set_ylabel(f'Modified $^{{207}}$Pb age ({wav_plot_prefix})')
        results['fig_wav'] = fig

    return results


def fc_A48i(fit57, fit86, A, sA, t0=1.0, norm_isotope='204Pb',
            negative_ar=True, negative_ages=True, hist=(False, False),
            trials=50_000):
    """
    Compute "forced concordance" [234U/238U] value following the approach of
    Engel et al., (2019).

    Parameters
    ----------
    fit57 :
        207Pb isochron linear regression fit
    fit86 :
        206Pb isochron linear regression fit
    A : array-like
        one-dimensional array of activity ratio values arranged as follows
        - np.nan, [230Th/238U], [226Ra/238U], [231Pa/235U]
    sA : array-like
        one-dimensional array of activity ratio value uncertainties given
        as 1 sigma absolute and arranged in the same order as A

    """

    # compute iso-57 diseq age
    t57 = isochron(fit57['theta'][1], A[-1], t0, 'iso-Pb7U5')
    # compute concordant init [234U/238U] value
    DC8 = (cfg.lam238, cfg.lam234, cfg.lam230, cfg.lam226)
    BC8 = ludwig.bateman(DC8)

    A48i = concordant_A48i(t57, fit86['theta'][1], A[1], A[2], DC8, BC8,
                           A48i_guess=1.)

    # Compute init [234U/238U] uncertainties:
    mc = dqmc.fc_A48i(t57, A48i, fit57, fit86, A, sA, [True, True], trials=trials,
                    negative_ar=negative_ar, negative_ages=negative_ages,
                    hist=hist)

    results = {
        'norm_isotope': norm_isotope,
        '207Pb_age': t57,
        '207Pb_age_95pm': [0., 0.],
        '[234U/238U]i': A48i,
        'mc': mc,
    }
    return results


#=======================================
# Numerical age calculation functions
#=======================================

def concint(a, b, A, init, t0):
    """
    Numercially compute disequilibrium U-Pb concordia-intercept age.
    """
    DC8 = (cfg.lam238, cfg.lam234, cfg.lam230, cfg.lam226)
    DC5 = (cfg.lam235, cfg.lam231)
    BC8 = ludwig.bateman(DC8)
    BC5 = ludwig.bateman(DC5, series='235U')
    fmin, dfmin = minimise.concint(t0=1.0, diagram='tw', init=init)
    args = (a, b, A[:-1], A[-1], DC8, DC5, BC8, BC5, cfg.U)
    r = optimize.newton(fmin, t0, dfmin, full_output=True, disp=False, args=args)
    if not r[1].converged:
        raise exceptions.ConvergenceError('disequilibrium concordia age did '
                  'not converge after maximum number of iterations')
    t = r[0]
    return t


def isochron(b, A, t0, age_type, init=(True, True)):
    """
    Numerically compute disequilbrium U-Pb isohcron age.
    """
    assert age_type in ('iso-Pb6U8', 'iso-Pb7U5')
    if age_type == 'iso-Pb6U8':
        DC = (cfg.lam238, cfg.lam234, cfg.lam230, cfg.lam226)
        BC = ludwig.bateman(DC)
    elif age_type == 'iso-Pb7U5':
        DC = (cfg.lam235, cfg.lam231)
        BC = ludwig.bateman(DC, series='235U')
    args = (b, A, DC, BC)
    fmin, dfmin = minimise.isochron(t0=t0, age_type=age_type, init=init)
    r = optimize.newton(fmin, t0, dfmin, full_output=True, disp=False,
                        args=args)
    if not r[1].converged:
        raise exceptions.ConvergenceError(f'disequilibrium isochron age did not '
              f'converge did not converge after maximum number of iterations, '
              f'fmin = {fmin(r[0], *args)}')
    t = r[0]
    return t


def pbu(x, A, t0, age_type, init=(True, True)):
    """
    Numerically compute single analysis Pb/U age using Ludwig equations.
    """
    assert age_type in ('Pb6U8', 'Pb7U5')
    if age_type == 'Pb6U8':
        DC = (cfg.lam238, cfg.lam234, cfg.lam230, cfg.lam226)
        BC = ludwig.bateman(DC)
    elif age_type == 'Pb7U5':
        DC = (cfg.lam235, cfg.lam231)
        BC = ludwig.bateman(DC, series='235U')
    args = (x, A, DC, BC)
    fmin, dfmin = minimise.pbu(t0=t0, age_type=age_type, init=init)
    t, r = optimize.newton(fmin, t0, dfmin, full_output=True, disp=False,
            args=args)
    if not r.converged:
        raise exceptions.ConvergenceError('Pb/U age routine did not converge '
                'after maximum number of iterations')
    return t


def mod207(x, y, A, Pb76, t0, init=(True, True)):
    """
    Numerically compute modified 207Pb ages using the equations of Ludwig (1977).
    """
    DC8 = (cfg.lam238, cfg.lam234, cfg.lam230, cfg.lam226)
    DC5 = (cfg.lam235, cfg.lam231)
    BC8 = ludwig.bateman(DC8)
    BC5 = ludwig.bateman(DC5, series='235U')
    args = (x, y, A, DC8, DC5, BC8, BC5, cfg.U, Pb76)
    fmin, dfmin = minimise.mod207(t0=t0, init=init)
    t, r = optimize.newton(fmin, t0, dfmin, full_output=True, args=args)
    if not r.converged:
        raise exceptions.ConvergenceError('Modified 207Pb age routine did not '
                'converge after maximum number of iterations')
    return t


def guillong(x, fXU, t0, age_type):
    """
    Numerically compute  single analysis Pb/U age using the equations and
    approach of Guillong (2014).

    Parameters
    ----------
    x : float or np.ndarray
        measured Pb/U ratio(S)
    fXU : float of np.ndaraay
        Th-U or Pa-U "fractionation factor"
    """
    if age_type == 'Pb6U8':
        DC = (cfg.lam238, cfg.lam230)
    else:
        DC = (cfg.lam235, cfg.lam231)
    fmin, dfmin = minimise.guillong(t0=t0)
    t, nr = optimize.newton(fmin, t0, dfmin, args=(x, fXU, *DC), full_output=True)
    if not nr.converged:
        raise exceptions.ConvergenceError('Guillong modified Pb/U age routine '
                  'did not converge after maximum number of iterations')
    return t


def sakata(x, y, fThU, fPaU, Pb76, t0, disp=False):
    """
    Numerically compute modified 207Pb age using the equations and approach of
    Sakata (2017).
    """
    fmin, _ = minimise.sakata(t0=t0)
    args = (x, y, fThU, fPaU, cfg.lam238, cfg.lam230, cfg.lam235, cfg.lam231,
            cfg.U, Pb76)
    t, nr = optimize.bisect(fmin, 1e-06, 100, args=args, full_output=True,
                            disp=disp)
    if not nr.converged:
        raise exceptions.ConvergenceError('Sakata modified 207Pb age routine '
                  'did not converge after maximum number of iterations')
    return t


def concordant_A48i(t75, b86, A08, A68, DC8, BC8, A48i_guess=1.):
    """
    Numerically compute initial U234/U238 activity ratio that forces concordance
    between 238U and 235U isochron ages.

    Minimises function: f = F(t75, A234A238) - slope_86, where t75 is
    the 207Pb/x-235U/x isochron age.
    """
    args = (t75, b86, [nan, A08, A68], DC8, BC8)
    fmin, dfmin = minimise.concordant_A48()
    r = optimize.newton(fmin, A48i_guess, dfmin, args=args,
            full_output=True, disp=False)
    if not r[1].converged:
        raise exceptions.ConvergenceError('forced concordant initial [234U/238U] '
               'routine did not converge after maximum number of iterations')
    return r[0]


def concint_multiple(a, b, A, init, t0, age_lim=(0., 20.), t_step=1e-5,
        A48i_lim=(0., 20.), A08i_lim=(0., 20.)):
    """
    Search for all initial disequilibrium concordia-intercept age solutions
    within specified age and initial activity ratio limits.

    If both [234U/238U] and [230Th/238U] are inputted as present-day
    values, it is possible for the concordia curve to curve around leading to two
    concordia intercepts in close age proximity. Typically, where this is the 
    case, it seems the  upper intercept will have a physically implausible 
    initial activity ratio solution, so lower intercept is chosen.

    """
    if not age_lim[0] <= t0 <= age_lim[1]:
        raise ValueError('age guess must be b/w upper and lower age limits')

    DC8 = (cfg.lam238, cfg.lam234, cfg.lam230, cfg.lam226)
    DC5 = (cfg.lam235, cfg.lam231)
    BC8 = ludwig.bateman(DC8)
    BC5 = ludwig.bateman(DC5, series='235U')

    # compile args for age solution
    args = (a, b, A[:-1], A[-1], DC8, DC5, BC8, BC5, cfg.U)

    # Find all solutions within age limits -- in case there are more than 1.
    fmin, dfmin = minimise.concint(t0, diagram='tw', init=init)
    roots, _ = find_roots(fmin, dfmin, args, range=age_lim, step=t_step)

    # If no ages found yet, try using inputted age guess directly in numerical
    # age routine:
    if len(roots) == 0:
        try:
            t = concint(a, b, A, init, t0)
        except exceptions.ConvergenceError:
            raise RuntimeError('no disequilibrium age solutions found')
        roots = np.atleast_1d(t)

    # Reject ages outside limits
    accept = np.where(np.logical_and(age_lim[0] < roots, roots < age_lim[1]),
                      np.full(roots.shape, True), np.full(roots.shape, False))
    accept = np.atleast_1d(accept)
    roots = roots[accept]

    # User youngest intercept age
    if roots.size > 0:
        ind = np.argsort(roots)
        roots = roots[ind]

    # Calculate activity ratios if present-day ratios given
    A48i = np.zeros(roots.shape) if not init[0] else None
    A08i = np.zeros(roots.shape) if not init[1] else None

    for i, t in enumerate(roots):
        if not init[0]:
            # present 234U/238U activity ratio:
            A48i[i] = useries.ar48i(t, A[0], cfg.lam238, cfg.lam234)
            if not init[1]:
                A08i[i] = useries.ar08i(t, A[0], A[1], cfg.lam238, cfg.lam234,
                                        cfg.lam230, init=init[0])
        elif not init[0]:
            # initial 234U/238U activity ratio:
            A08i[i] = useries.ar08i(t, A[0], A[1], cfg.lam238, cfg.lam234,
                                    cfg.lam230, init=init[0])

    # check that activiy ratios are within limits
    if not init[0]:
        accept = np.where(np.logical_and(A48i_lim[0] < A48i,
                        A48i < A48i_lim[1]), accept,
                        np.full(roots.shape, False))
    if not init[1]:
        accept = np.where(np.logical_and(A08i_lim[0] < A08i,
                        A08i < A08i_lim[1]), accept,
                        np.full(roots.shape, False))

    # now mask out rejected values
    ages = roots[accept]
    if A48i is not None:
        A48i = A48i[accept]
    if A08i is not None:
        A08i = A08i[accept]

    if len(ages) < 1:
        raise ValueError('no disequilibrium age solutions found within '
                         'age and activity ratio limits')

    return ages, A48i, A08i


def find_roots(f, df, args, range=(0., 20.), step= 1e-5, rtol=1e-9):
    """
    Use brute force method to find all roots within given age range. Not
    guaranteed to work if two roots are close together, so use small step
    size.
    """
    # Find where f changes sign
    x = np.arange(range[0], range[1] + step, step)
    y = f(x, *args)
    sign_change = np.diff(np.sign(y)) != 0.
    x_ind = np.where(sign_change)[0]

    # Find guess at roots
    x0 = x[x_ind]
    roots = []

    # Polish roots using Newton-Raphson
    for i, v in enumerate(x0):
        root, rr = optimize.newton(f, v, df, args=args, rtol=rtol,
                                   full_output=True, disp=False)
        if rr.converged and not np.isnan(root):
            roots.append(root)

    # Remove any duplicate roots
    if len(roots) != 0:
        out = [roots[0]]
        for r in roots[1:]:
            if not np.any(np.isclose(np.full(len(out), r),
                                     out, atol=0, rtol=1e-04)):
                out.append(r)
    else:
        out = []

    return np.array(out), y


#=======================================
# Analytical age uncertainty functions
#=======================================

# def conint_age_uncert():
#     """ Analytical concordia-intercept age uncertainties.
#     """
#     return


# def mod207Pb_age_uncert():
#     """ Analytical mod. 207Pb age uncertainties.
#     """
#     return








"""
Linear regression algorithms for 2-dimensional datasets

References
----------
.. [Faure1977]
    Faure, G., 1977. Appendix 1: Fitting of isochrons for dating by the Rb-Sr
    method, in: Principles of Isotope Geology. John Wiley and Sons, pp. 1–17.
.. [Ludwig2012]
    Ludwig, K.R., 2012. Isoplot/Ex Version 3.75: A Geochronological Toolkit for
    Microsoft Excel, Special Publication 4. Berkeley Geochronology Center.
.. [Ludwig1980]
    Ludwig, K.R., 1980. Calculation of uncertainties of U-Pb isotope data. Earth
    and Planetary Science Letters 212–202.
.. [Powell2020]
    Powell, R., Green, E.C.R., Marillo Sialer, E., Woodhead, J., 2020. Robust
    isochron calculation. Geochronology 2, 325–342.
    https://doi.org/10.5194/gchron-2-325-2020
.. [Sadovski1974]
    Sadovski, A.N., 1974. Algorithm AS 74: L1-norm Fit of a Straight Line. Journal
    of the Royal Statistical Society. Series C (Applied Statistics) 23,
    244–248. https://doi.org/10.2307/2347013
.. [Siegel1982]
    Siegel, A. F.: Robust regression Using repeated medians, Biometrika,
    69, 242–244, 1982.
.. [Titt1979]
    Titterington, D.M., Halliday, A.N., 1979. On the fitting of parallel
    isochrons and the method of maximum likelihood. Chemical Geology 26, 183–195.
.. [York2004]
    York, D., Evensen, N.M., Martı́nez, M.L., De Basabe Delgado, J., 2004. Unified
    equations for the slope, intercept, and standard errors of the best
    straight line. American Journal of Physics 72, 367–375.
    https://doi.org/10.1119/1.1632486

"""

import numpy as np

# from matplotlib.offsetbox import AnchoredText

from . import cfg
from . import exceptions
from . import stats
from . import plotting


#=================================
# Regression fitting routines
#=================================

def classical_fit(x, sx, y, sy, r_xy, model='ca', plot=False, diagram=None,
                  isochron=True, xlim=(None, None), ylim=(None, None),
                  axis_labels=(None, None), norm_isotope=None, dp_labels=None):
    """
    Fit a classical regression line to data and optionally create plot of the
    fit. If model is set to 'ca', then routine will emulate default the
    behaviour of Isoplot (Ludwig, 2012) .

    Parameters
    -----------
    x, sx, y, sy, r_xy : np.ndarray
        Data points and error correlations as 1d arrays.
    isochron : bool, optional
        If true and model is 'ca' then fits model 3 for excess scatter data
        sets instead of model 2.
    model: str, optional
        'ca': fit 'best' model depending on MSWD of york fit; emulates Isoplot
                behaviour
        'c1': standard York fit with analytical errors
        'c2': model of McSaveney in Faure (1977), equivalent to Isoplot model 2
        'c3': equivalent to Isoplot model 3
    """
    assert model in ('ca', 'c1', 'c2', 'c3')

    n = x.size
    if not n == sx.size == y.size == sy.size == r_xy.size:
        raise ValueError('incompatible number of elements in input data arrays')
    if n == 1:
        raise ValueError('cannot fit linear regression to a single data point')

    df = n - 2
    sy_excess = None

    # fit York model:
    theta, covtheta, mswd, xbar, ybar, r = york(x, sx, y, sy, r_xy)

    # get pr fit:
    p = stats.pr_fit(df, mswd)

    # get mswd_regression_thresholds:
    mswd_threshold_1 = stats.mswd_conf_limits(df, p=cfg.mswd_ci_thresholds[0])
    mswd_threshold_2 = stats.mswd_conf_limits(df, p=cfg.mswd_ci_thresholds[1])

    # equivalent p thresholds -
    p_threshold_1 = 1. - cfg.mswd_ci_thresholds[0]
    p_threshold_2 = 1. - cfg.mswd_ci_thresholds[1]

    if model == 'c1' or (model == 'ca' and mswd < mswd_threshold_2):
        if mswd < mswd_threshold_1:
            t_crit = 1.96
            theta_95ci = t_crit * np.sqrt(np.diag(covtheta))
            fitted_model = 'model 1'
            excess_scatter = False
        else:
            t_crit = stats.t_critical(df)
            covtheta *= mswd
            theta_95ci = t_crit * np.sqrt(np.diag(covtheta))
            fitted_model = 'model 1x'
            excess_scatter = True

    elif model == 'c2' or (model == 'ca' and not isochron):
        m2 = york(x, sx, y, sy, r_xy, model='2')
        theta, covtheta, mswd2, xbar, ybar, r = m2
        t_crit = stats.t_critical(df)
        covtheta *= mswd2
        theta_95ci = t_crit * np.sqrt(np.diag(covtheta))
        fitted_model = 'model 2'
        excess_scatter = True

    else:
        t_crit = stats.t_critical(df)
        sy_excess0 = np.sqrt(mswd) * np.sqrt(np.diag(covtheta))[0]  # use including scattter error
        fit_results = model_3(x, sx, y, sy, r_xy, covtheta, sy_excess0,
                              theta0=theta)
        (theta, covtheta, mswd3, xbar, ybar, r, sy_excess) = fit_results
        fitted_model = 'model 3'
        excess_scatter = True
        theta_95ci = t_crit * np.sqrt(np.diag(covtheta))

    fit = {
        'type': 'classical',
        'model': fitted_model,
        'excess_scatter': excess_scatter,
        'n': n,
        'sy_excess_1s': sy_excess,
        'theta': theta,
        'covtheta': covtheta,
        'theta_95ci': theta_95ci,
        't_crit': t_crit,
        'r_ab': covtheta[1, 0] / np.prod(np.sqrt(np.diag(covtheta))),
        'wtd_residuals': r,
        'mswd':  mswd,
        'mswd_regression_thresholds': [mswd_threshold_1, mswd_threshold_2],
        'p': p,
        'p_thresholds': [p_threshold_1, p_threshold_2],
        'xbar': xbar,
        'ybar': ybar
    }

    if plot:
        fig = plotting.plot_dp(x, sx, y, sy, r_xy, labels=dp_labels)
        plotting.apply_plot_settings(fig, plot_type='isochron', diagram=diagram,
                xlim=xlim, ylim=ylim, axis_labels=axis_labels,
                norm_isotope=norm_isotope)
        plotting.plot_rfit(fig.axes[0], fit)
        fit['fig'] = fig

    return fit


def robust_fit(x, sx, y, sy, r_xy, model='ra', plot=False,
        diagram=None, xlim=(None, None), ylim=(None, None),
        axis_labels=(None, None), norm_isotope='204Pb', dp_labels=None):
    """
    Fit data with robust regression line and optionally create a plot of
    the fit.

    Parameters
    ----------
    x, sx, y, sy, r_xy : np.ndarray
        Data points and error correlations as 1d arrays.
    model: str, optional
        ra : fit 'best' robust model depending on whether or no spine width
            is < upper limit
        rs : spine fit of Powell et al., (2020)
        r2 : robust model 2, robust version of the Isoplot model 2
        rx : spine fit with expanded errors, for use if s is a little bit over
            slim (experimental feature only).

    """
    assert model in ('ra', 'rs', 'r2', 'rx')

    n = x.size
    if not n == sx.size == y.size == sy.size == r_xy.size:
        raise ValueError('incompatible number of elements in input data arrays')

    excess_scatter = False

    # fit spine model:
    if model in ('ra', 'rs', 'rx'):
        theta, covtheta, r, code = spine(x, sx, y, sy, r_xy, h=cfg.h)
        s = stats.nmad(r)
        s_upper95 = slim(n)
        excess_scatter = True if s > s_upper95 else False
        fitted_model = 'spine'

        if 'rx':    # experimental only
            if excess_scatter:
                covtheta *= (s / s_upper95) ** 2
                fitted_model = 'spine x'

    if model == 'r2' or (model == 'ra' and excess_scatter):
        try:
            theta, covtheta, r = robust_model_2(x, y, xy_err=[sx, sy, r_xy])
        except exceptions.ConvergenceError:
            raise RuntimeError('robust model 2 fit failed to converge')

        fitted_model = 'robust model 2'

        # If no spine fit, then set s to nan
        if model == 'r2':
            s = np.nan
            s_upper95 = np.nan
            excess_scatter = np.nan

    theta_95ci = 1.96 * np.sqrt(np.diag(covtheta))

    fit = {
        'type': 'robust',
        'model': fitted_model,
        'excess_scatter': excess_scatter,
        'n': n,
        'theta': theta,
        'covtheta': covtheta,
        'theta_95ci': theta_95ci,
        'r_ab': covtheta[1, 0] / np.prod(np.sqrt(np.diag(covtheta))),
        't_crit': 1.96,
        'wtd_residuals': r,
        's': s,
        'slim': s_upper95
    }

    if plot:
        fig = plotting.plot_dp(x, sx, y, sy, r_xy, labels=dp_labels)
        plotting.apply_plot_settings(fig, plot_type='isochron', diagram=diagram,
                xlim=xlim, ylim=ylim, axis_labels=axis_labels,
                norm_isotope=norm_isotope)
        plotting.plot_rfit(fig.axes[0], fit)
        fit['fig'] = fig

        # May add text boxes in future...
        # ax = fig.axes[0]
        #
        # text = f"slope: -0.4735 ± 0.079 e-05 " \
        #        f"\n y-int. 0.8137 ± 0.00077" \
        #        f"\n $s$: 1.14 (1.48, $n$ = {fit['n']})"
        # at = AnchoredText(text, prop=dict(size=7), frameon=True,
        #         loc='upper right')
        # at.patch.set_boxstyle("round,pad=0.,rounding_size=0.5")
        # at.patch.set_linewidth(0.75)
        # ax.add_artist(at)

    return fit


#=================================
# Basic line fitting functions
#=================================

def lsq(x, y):
    """
    Fit ordinary least-squares model
    Based on code by Roger Powell.
    """
    n = x.shape[0]
    x = np.column_stack((np.ones(n), x))
    y = y.reshape(n, 1)
    inv = np.linalg.inv(x.T @ x)
    theta = inv @ x.T @ y
    e = np.dot(x, theta) - y
    sigfit2 = (e.T @ e) / (n - 2)
    return theta, sigfit2 * inv


def lad(x, y):
    """
    Fit least absolute deviation model of Sadovski (1974).
    Based on code by Roger Powell.
    """
    n = x.size
    rr = 1e-8 * cfg.rng.random(n - 1) # used for naive breaking of x ties
    bi = np.empty(n)
    bi.fill(False)
    k, i, i1, i2 = 0, 0, 0, 0

    while i != -1 and k < 12:
        i2 = i1
        i1 = i
        k += 1

        o = np.delete(np.arange(n), i)
        x1 = np.delete(x - x[i], i) + rr
        y1 = np.delete(y - y[i], i)
        oo = np.argsort(y1 / x1)
        x2 = np.abs(x1[oo])
        mid = sum(x2)/2
        sx = 0
        j = 0

        while sx < mid:
            sx += x2[j]
            j += 1
            i = o[oo][j-1]
            if i == i2:
                i = -1
            elif bi[i]:
                i = -1
            else:
                bi[i] = True

    theta = np.array(((x[i1] * y[i2] - x[i2] * y[i1])/(x[i1] - x[i2]),
                     (y[i1] - y[i2])/(x[i1] - x[i2])))
    return theta


def siegel(x, y):
    """
    Median of pairwise median slopes algorithm of Siegel (1982).
    Based on code by Roger Powell.
    """
    n = x.size
    x = x + 1e-8 * cfg.rng.random(n)  # naive breaking of x ties

    med = np.empty(n)
    for i in range(n):
        pmed = np.empty(n)
        for j in range(n):
            if i is not j:
                pmed[j] = (y[j] - y[i]) / (x[j] - x[i])
        med[i] = np.median(np.delete(pmed, i))

    b = np.median(med)
    theta = np.array((np.median(y - x * b), b))
    return theta


#=================================
# Classical stats algorithms
#=================================

def york(x, sx, y, sy, r_xy, sy_excess=None, model="1", itmax=50, rtol=1e-08,
         atol=1e-08, theta0=None):
    """
    Fit total weighted least squares regression using the algorithm
    of York et al. (2004). Analytical errors on regression
    parameters are equivalent to the Maximum Likelihood approach of
    Titterington and Halliday (1979).

    Can also fit a version of the Isoplot model 2 and model 3 by adjusting the
    weightings. Model 2 code is based on McSaveney in Faure (1977).
    Model 3 code emulates Isoplot (Ludwig, 2012).

    Notes
    -----
    Model 3 fits involve an iterative approach that typically requires calling
    this function several times. To perform a model 3 fit, call the model_3
    function.

    """
    assert model in ("1", "2", "3")
    if model == '3' and sy_excess is None:
        raise ValueError('inital sy_excess value must provided for morel 3 fit')

    n = x.shape[0]

    # centre data
    avx = np.mean(x)
    avy = np.mean(y)
    div = np.array([1 / avy, avx / avy])
    x = x / avx
    sx = sx / avx
    y = y / avy
    sy = sy / avy

    if sy_excess is not None:
        sy_excess /= avy

    # break ties
    # TODO: is this necessary?
    x += 1e-9 * cfg.rng.random(n)

    if theta0 is None:
        theta0 = siegel(x, y)
    # b0 = theta0[1]
    b0 = (theta0 * div)[1]               # equivalent slope for centred data

    if model == "1":
        # Weight each point according to analytical uncertainties.
        wx = 1. / sx ** 2
        wy = 1. / sy ** 2
        cor = r_xy
    elif model == "2":
        # Weight each point equally with zero correlation - wy is assigned in
        # main for loop.
        wx = np.ones(n)
        cor = np.zeros(n)
    else:  # Model 3
        # analytical errors only for x
        wx = 1. / sx ** 2
        # combined analytical and initial ratio errors for y
        wy = 1. / (sy ** 2 + sy_excess ** 2)
        cor = r_xy * np.sqrt(sy ** 2 / (sy ** 2 + sy_excess ** 2))

    i = 0
    while i < itmax:
        i += 1

        if model == "2":
            wy = np.ones(n) / b0 ** 2

        # Evaluate w
        alpha = np.sqrt(wx * wy)
        w = wx * wy / (wx + wy * b0 ** 2 - 2 * b0 * cor * alpha)

        # Calculate xbar, ybar, U, V and beta
        xbar = np.sum(w * x) / np.sum(w)
        ybar = np.sum(w * y) / np.sum(w)
        u = x - xbar
        v = y - ybar
        beta = w * (u / wy + b0 * v / wx - (b0 * u + v) * cor / alpha)

        # Find improved b
        b = np.sum(w * beta * v) / np.sum(w * beta * u)

        # Check for convergence.
        if np.isclose(b0, b, atol=atol, rtol=rtol):

            # goodness of fit
            a = ybar - b * xbar
            theta = np.array((a, b))
            # s = np.sum(w * (y - b * x - a) ** 2)
            r = np.sqrt(w) * (y - b * x - a)
            s = np.sum(r ** 2)
            mswd = s / (n - 2) if n > 2 else np.nan

            # get analytical errors following York et al., (2004)
            xadj = xbar + beta
            xadj_bar = np.sum(w * xadj) / np.sum(w)
            uadj = xadj - xadj_bar
            sb = np.sqrt(1. / np.sum(w * uadj ** 2))
            sa = np.sqrt(1. / np.sum(w) + xadj_bar ** 2 * sb ** 2)
            r_ab = -xadj_bar / np.sqrt(np.sum(w * xadj ** 2) / np.sum(w))
            covtheta = np.array([[sa ** 2, r_ab * sb * sa],
                                 [r_ab * sa * sb, sb ** 2]])

            # # analytical errors following Powell et al., (2002; 2020)
            # #   - matrix approach
            # X = np.array((np.ones(n), x)).T
            # covtheta = np.linalg.inv(X.T @ np.diag(w) @ X)

            # 'de-centre' results
            theta /= div
            covtheta /= np.array([[div[0] ** 2, div[0] * div[1]],
                                  [div[0] * div[1], div[1] ** 2]])
            xbar *= avx
            ybar *= avy

            return theta, covtheta, mswd, xbar, ybar, r

        # Update estimate of b0
        b0 = b

    raise exceptions.ConvergenceError("york regression routine did not "
                "converge after maximum number of iterations")


def model_2(x, y):
    """
    Fit a model 2 regression line. The slope is the geometric mean of ordinary
    least squares of y on x and the reciprocal of the ordinary least squares
    of x on y (Powell et al., 2020).

    Equivalent to Isoplot Model 2 fit which is based on the routine of
    MacSaveney in Faure (1977) and uses the main York routine with weightings
    set to: wx = 0., wy = 1 / b^2, p_xy = 0, to achieve the same result
    iteratively.

    Notes
    -----
    Use the york function for now with model set to "2" as this function does
    not yet compute covtheta.

    """
    n = x.shape[0]
    xbar = (x @ np.ones(n)) / n
    ybar = (y @ np.ones(n)) / n
    u = x - xbar
    v = y - ybar
    b = np.sqrt(np.sum(v ** 2) / np.sum(u ** 2))

    # Check whether to take positive or negative root.
    sxy = np.sum(u * v)
    if sxy < 0:     # TODO: double check this
        b *= -1.

    a = ybar - b * xbar
    theta = np.array((a, b))

    return theta, xbar, ybar


def model_3(x, sx, y, sy, cor, covtheta, sy_excess0, theta0, mswd_tol=1e-04,
            itmax=100, york_itmax=50, york_atol=1e-08, york_rtol=1e-08):
    """
    Fir a model 3 regression line (equivalent to Isoplot model 3). Weights each
    point according to analytical errors, plus an extra component of Gaussian
    distributed scatter in y (applied equally to each data point). For each
    iteration this excess error in y is inflated and a new York fit produced,
    until MSWD = 1.

    In Isoplot, this is only used for classical isochron ages, but it could also
    be used for Tera-Wasserburg concordia-intercept ages (?). Whatever
    dataset it is applied to, there must be a good justification for the
    assumption of Guassian distributed geological scatter.

    """
    code = 0
    sy_excess = sy_excess0     # starting guess of excess scatter in y
    i = 0
    while i < itmax:
        i += 1
        try:
            theta, covtheta, mswd, xbar, ybar, r = york(x, sx, y, sy, cor,
                            theta0=theta0, sy_excess=sy_excess, model="3",
                            itmax=york_itmax, rtol=york_rtol, atol=york_atol)
            if mswd < 1.:
                raise ValueError(f'model 3 fit failed - MSWD is < 1 after '
                                 f'iteration {i}')

        except exceptions.ConvergenceError:
            raise exceptions.ConvergenceError(f'york failed to converge for '
                       f'iteration {i} with mswd = {mswd}')

        if abs(mswd - 1) < mswd_tol:    # check convergence
            return theta, covtheta, mswd, xbar, ybar, r, sy_excess

        # update guess of excess scatter in y
        sy_excess *= np.sqrt(mswd)
        theta0 = theta

    raise exceptions.ConvergenceError(f' 3 routine failed - mswd (= {mswd}) '
              f'did not succesfully converge to 1')


#=================================
# Robust algorithms
#=================================

def slim(n):
    """Upper bound of 95% confidence interval on s (spine width) for Spine
    linear regression algorithm. Derived from simulation of Gaussian distributed
    datasets. See Powell et al., (2020).
    """
    return 1.92 - 0.162 * np.log(10 + n)


def spine(x, sx, y, sy, cor, h=1.4):
    """
    Iteratively re-weighted Huber line-fitting algorithm of Powell et al.,
    (2020).

    Code by Roger Powell with style modifications by TP.

    Notes
    ------
    Updated to use normalised deltheta as convergence criteria for improved
    performance where slope and y-int are of very different order of
    magnitude.

    """
    n = x.shape[0]
    itmax = 40
    mindel = 5e-7
    mincond = 1e-12

    # centre data
    avx = np.dot(x, np.ones(n)) / n
    avy = np.dot(y, np.ones(n)) / n
    div = np.array([1 / avy, avx/avy])
    x = x / avx
    sx = sx / avx
    y = y / avy
    sy = sy / avy
    cov = sx * sy * cor

    theta = ntheta = siegel(x, y)

    i = 0
    code = 0
    ndeltheta = (1e10, 1e10)

    while i < itmax:
        i += 1
        a, b = oldtheta = theta
        e = a + b * x - y
        sde = np.sqrt(b ** 2 * sx ** 2 - 2 * b * cov + sy ** 2)
        r = e / sde
        wh = [1 if abs(rk) < h else np.sqrt(h/abs(rk)) for rk in r] / sde

        xp = x - r * (b * sx ** 2 - cov) / sde                 # x on the line
        ypp = (y - e - r * (b * cov - sy ** 2) / sde) * wh     # W^(1/2)(y'-e): y off the line
        c = np.transpose([wh, xp * wh])                        # W^(1/2) X'
        (u, s, v) = np.linalg.svd(c, full_matrices=False)

        if mincond * s[0] > s[1]:
            raise exceptions.ConvergenceError('spine algorithm did not converge - (nearly) '
                           'singular matrix')

        theta = np.dot(np.dot(np.dot(np.transpose(v), np.diag(1/s)),
                              np.transpose(u)), ypp)
        ndeltheta = (theta - oldtheta) / ntheta

        if np.sqrt(np.dot(ndeltheta, ndeltheta)) < mindel:
            dpsi = [1 if abs(rk) < h else 0 for rk in r]
            denom = np.array([[div[0] ** 2, div[0] * div[1]],
                              [div[0] * div[1], div[1] ** 2]])
            covtheta = np.linalg.inv(c.T @ np.diag(dpsi) @ c) / denom
            return theta / div, covtheta, r, code

    # not converged
    raise exceptions.ConvergenceError(f'spine linear regression routine did not '
               f'converge after maximum number of iterations - ndeltheta = '
               f'{ndeltheta}')


def huberu(x, y, xy_err=None, h=1.4, itmax=250):
    """
    Huber line-fitter with analytical errors discarded. Used for robust model
    2 fit. Can take many iterations to converge, so set itmax quite high.

    Original code by Roger Powell with some style modifications by TP.

    Parameters
    ----------
    xy_err : np.ndarray, optional
        3 x n array-like with elements sx, sy, r_xy (i.e. analytical
        errors on x and y, and their correlation coefficient). Used
        to ensure fit error is not less than average analytical error
        following approach of Powell et al. (2002).

    Notes
    -----
    Do not call this function directly, instead call robust_model_2.

    """
    n = x.shape[0]
    mindel = 1e-7
    mincond = 1e-12

    if xy_err is not None:
        sx = xy_err[0]
        sy = xy_err[1]
        r_xy = xy_err[2]

    i = 0
    ndeltheta = (1e10, 1e10)
    theta = ntheta = siegel(x, y)

    while i < itmax:
        i += 1
        (a, b) = oldtheta = theta
        e = a + b * x - y
        scat = stats.nmad(e)

        # Check if scat is smaller than average analytical uncertainties,
        # and if so, reset average analyutical uncertainties.
        if xy_err is not None:
            se = np.sqrt(sy ** 2 + (b * sx) ** 2 - 2 * b * sx * sy * r_xy)
            averr = np.median(se)
            scat = averr if averr > scat else scat

        r = e / scat
        wh = [1 if abs(rk) < h else np.sqrt(h / abs(rk)) for rk in r] / scat
        xp = x                                           # x on the line
        ypp = y * wh                                     # W^(1/2)(y'-e): y off the line
        c = np.transpose([wh, xp * wh])                  # W^(1/2) X'
        u, s, v = np.linalg.svd(c, full_matrices=False)

        if mincond * s[0] > s[1]:
            raise exceptions.ConvergenceError('huberu algorithm failed: '
                      '(nearly) singular matrix')

        theta = v.T @ np.diag(1 / s) @ u.T @ ypp
        ndeltheta = (theta - oldtheta) / ntheta

        if np.sqrt(np.dot(ndeltheta, ndeltheta)) < mindel:
            dpsi = [1 if abs(rk) < h else 0 for rk in r]
            covtheta = np.linalg.inv(c.T @ np.diag(dpsi) @ c)
            return theta, covtheta

    raise exceptions.ConvergenceError('huberu algorithm did not converge after '
              'maximum number of iterations')


def robust_model_2(x, y, xy_err=None, h=1.4, disp=True):
    """
    Fit robust model 2 regression line.

    Analytical errors discarded and takes the geometric mean slope of the spine
    y on x fit, and the reciprocal of the spine x on y fit, using data scatter
    in place of analytical weights. Similar to the Isoplot model 2 approach
    (Ludwig, 2012) but with robust properties.

    If xy_err is provided, nmad of the residuals will be checked against the
    median analytical uncertainty, and if larger, this value will be used to
    weight data points instead. Following Powell et al., (2002), this is to
    guard against a situation in which the scatter on the data is smaller than
    that expected with the analytical uncertainties.

    Code by Roger Powell with style modifications by TP.

    Parameters
    -----------
    dp_err : np.ndarray, optional
        3 x n array-like with elements sx, sy, r_xy (i.e. analytical
            errors on x and y, and their correlation coefficient)

    """
    # centre data
    avx = np.mean(x)
    avy = np.mean(y)
    div = np.array([1 / avy, avx / avy])
    x = x / avx
    y = y / avy

    # y on x regression
    if xy_err is not None:
        err = [xy_err[0] / avx, xy_err[1] / avy, xy_err[2]]
    else:
        err = None

    thyx, covthetayx, = huberu(x, y, xy_err=err, h=h)
    ayx, byx = thyx

    # x on y regression
    if xy_err is not None:
        err = [xy_err[1] / avy, xy_err[0] / avx, xy_err[2]]
    else:
        err = None

    thxy, covthxy = huberu(y, x, xy_err=err, h=h)
    axy, bxy = thetaxy = -thxy[0] / thxy[1], 1 / thxy[1]
    j = -bxy * np.array([[1, axy], [0, bxy]])
    covthetaxy = j @ covthxy @ np.transpose(j)

    # take geometric mean
    if byx > 0:
        avb = np.sqrt(byx * bxy)
    else:
        avb = -np.sqrt(byx * bxy)  # in general, care with signs
    ava = (axy * byx - ayx * bxy + avb * (ayx - axy)) / (byx - bxy)
    theta = (ava, avb)

    # residuals
    r = np.nan

    # This approximation is normally pretty good...
    # covavapprox = (covthetayx + covthetaxy) / 4

    jj = np.array([
        [bxy / (bxy + avb),
        -((-axy + ayx) * bxy * (bxy + byx - 2 * avb)) / (2 * avb * (-bxy + byx) ** 2),
         byx / (byx + avb),
         ((-axy + ayx) * byx * (bxy + byx - 2 * avb)) / (2 * avb * (-bxy + byx) ** 2)],
        [0, bxy / (2 * avb), 0, byx / (2 * avb)] ])

    cov = np.zeros((4, 4))
    cov[0, :2] = covthetayx[0]
    cov[1, :2] = covthetayx[1]
    cov[2, 2:] = covthetaxy[0]
    cov[3, 2:] = covthetaxy[1]

    covtheta = jj @ cov @ jj.T

    theta /= div
    covtheta /= np.array([[div[0] ** 2, div[0] * div[1]],
                          [div[0] * div[1], div[1] ** 2]])

    return theta, covtheta, r



"""
Disequilibrium Pb/U equations based on Ludwig (1977)

References
----------
.. [Ludwig1977]
    Ludwig, K.R., 1977. Effect of initial radioactive-daughter disequilibrium
    on U-Pb isotope apparent ages of young minerals. Journal of Research of
    the US Geological Survey 5, 663–667.

"""

import numpy as np

from .. import cfg
from .. import useries


exp = np.exp
log = np.log
nan = np.nan


# =================================
# Diseq Pb/U ratios
# =================================

def f(t, A, DC, BC, init=(True, True)):
    """
    206Pb*/238U ratio (where * denotes radiogenic Pb) as a function of t and
    activity ratio values following Ludwig (1977). Note there is a small typo
    in the original article.
    """
    # Get initial activity ratios if present value given.
    A48i = A[0] if init[0] else useries.ar48i(t, A[0], DC[0], DC[1])
    A08i = A[1] if init[1] else useries.ar08i(t, A[0], A[1], DC[0], DC[1], DC[2], init=True)
    
    # Components of F from 238U and each intermediate daughter
    # ignoring 210Pb
    F = exp(DC[0] * t) * (BC[0] * exp(-DC[0] * t) + BC[1] * exp(-DC[1] * t)
            + BC[2] * exp(-DC[2] * t) + BC[3] * exp(-DC[3] * t) + 1.)
    F += A48i*DC[0]/DC[1]*exp(DC[0]*t) * (BC[4] * exp(-DC[1] * t)
            + BC[5] * exp(-DC[2] * t) + BC[6] * exp(-DC[3] * t) + 1.)
    F += A08i*DC[0]/DC[2]*exp(DC[0]*t) * (BC[7] * exp(-DC[2] * t)
            + BC[8] * exp(-DC[3] * t) + 1.)
    F += A[2]*DC[0]/DC[3]*exp(DC[0]*t) * (1. - exp(-DC[3] * t))

    return F


def g(t, A15i, DC, BC):
    """
    207Pb*/235U (where * denotes radiogenic Pb) ratio as a function of t and
    activity ratio values following Ludwig (1977)
    """
    # ignoring 227Ac
    G = exp(DC[0] * t) * (BC[0] * exp(-DC[0] * t) + BC[1] * exp(-DC[1] * t) + 1.)
    G += A15i * DC[0] / DC[1] * exp(DC[0] * t) * (1. - exp(-DC[1] * t))
    return G


def f_comp(t, A, DC, BC, init=(True, True)):
    """
    Return individual components of f.
    """
    # Get initial activity ratios if present value given.
    A48i = A[0] if init[0] else useries.ar48i(t, A[0], DC[0], DC[1])
    A08i = A[1] if init[1] else useries.ar08i(t, A48i, A[1], DC[0], DC[1],
                                              DC[2], init=True)

    # Components of F from 238U and each intermediate daughter
        # ignoring 210Pb
    F1 = exp(DC[0] * t) * (BC[0] * exp(-DC[0] * t) + BC[1] * exp(-DC[1] * t)
            + BC[2] * exp(-DC[2] * t) + BC[3] * exp(-DC[3] * t) + 1.)
    F2 = A48i*DC[0]/DC[1]*exp(DC[0]*t) * (BC[4] * exp(-DC[1] * t)
            + BC[5] * exp(-DC[2] * t) + BC[6] * exp(-DC[3] * t) + 1.)
    F3 = A08i*DC[0]/DC[2]*exp(DC[0]*t) * (BC[7] * exp(-DC[2] * t)
            + BC[8] * exp(-DC[3] * t) + 1.)
    F4 = A[2]*DC[0]/DC[3]*exp(DC[0]*t) * (1. - exp(-DC[3] * t))

    return F1, F2, F3, F4


def g_comp(t, A15i, DC, BC):
    """
    Return individual components of g.
    """
    # ignoring 227Ac
    G1 = exp(DC[0] * t) * (BC[0] * exp(-DC[0] * t) + BC[1] * exp(-DC[1] * t) + 1.)
    G2 = A15i * DC[0] / DC[1] * exp(DC[0] * t) * (1. - exp(-DC[1] * t))
    return G1, G2


# =================================
# Bateman coefficients
# =================================

def bateman(DC, series='238U'):
    """
    Bateman coefficients for the 238U or 235U decay series as defined in
    Ludwig (1977) and elsewhere.

    DC : array-like
        array of relevant decay constants (see below)

    """
    assert series in ('238U', '235U')

    if series == '238U':

        # DC[0] = lam238
        # DC[1] = lam234
        # DC[2] = lam230
        # DC[3] = lam226
        # DC[4] = lam210

        # ignoring 210Pb
        c1 = DC[1] * DC[2] * DC[3] / ((DC[0] - DC[1]) * (DC[0] - DC[2]) * (DC[0] - DC[3]))
        c2 = DC[0] * DC[2] * DC[3] / ((DC[1] - DC[0]) * (DC[1] - DC[2]) * (DC[1] - DC[3]))
        c3 = DC[0] * DC[1] * DC[3] / ((DC[2] - DC[0]) * (DC[2] - DC[1]) * (DC[2] - DC[3]))
        c4 = DC[0] * DC[1] * DC[2] / ((DC[3] - DC[0]) * (DC[3] - DC[1]) * (DC[3] - DC[2]))

        h1 = -DC[2] * DC[3] / ((DC[1] - DC[2]) * (DC[1] - DC[3]))
        h2 = -DC[1] * DC[3] / ((DC[2] - DC[1]) * (DC[2] - DC[3]))
        h3 = -DC[1] * DC[2] / ((DC[3] - DC[1]) * (DC[3] - DC[2]))

        p1 = DC[3] / (DC[2] - DC[3])
        p2 = DC[2] / (DC[3] - DC[2])
        return np.array((c1, c2, c3, c4, h1, h2, h3, p1, p2))

    else:
        # DC[0] = lam235
        # DC[1] = lam231
        # DC[2] = lam227
        
        # ignoring 227Ac
        d1 = DC[1] / (DC[0] - DC[1])
        d2 = DC[0] / (DC[1] - DC[0])
        return np.array((d1, d2))



# =================================
# Derivatives of F and G
# =================================

def dfdt(t, A, init=(True, True), comp=False):
    """
    !!! Experimental function !!!
    Derivative of f with respect to t.

    Parameters
    ----------
    comp : bool
        if True, then returns individual components.

    """
    l8, l4, l0, l6 = cfg.lam238, cfg.lam234, cfg.lam230, cfg.lam226
    c1, c2, c3, c4, h1, h2, h3, p1, p2 = bateman((l8, l4, l0, l6))

    v2 = l8/l4 * (h1*exp((l8-l4)*t) + h2*exp((l8-l0)*t) + h3*exp((l8-l6)*t)
                  + exp(l8*t))
    dv2 = l8/l4 * (h1*(l8-l4)*exp((l8-l4)*t) + h2*(l8-l0)*exp((l8-l0)*t)
                     + h3*(l8-l6)*exp((l8-l6)*t) + l8*exp(l8*t))
    v3 = l8/l0 * (p1*exp((l8-l0)*t) + p2*exp((l8-l6)*t) + exp(l8*t))
    dv3 = l8/l0 * (p1*(l8-l0)*exp((l8-l0)*t) + p2*(l8-l6)*exp((l8-l6)*t)
                     + l8*exp(l8*t))

    if init[0]:  # iniital [234U/238]
        u2 = A[0]
        du2 = 0.
    else:  # present [234U/238U]
        u2 = (A[0] - 1.) * exp(l4 * t) + 1.
        du2 = l4 * (A[0] - 1.) * exp(l4 * t)

    if init[1]:  # initial [230Th/238U]
        u3 = A[1]
        du3 = 0.
    else:
        if init[0]:  # pres. [230Th/238U], initial [234U/238U]
            u3 = (A[0] - 1.)*exp(l0*t) - (A[0] - 1.) * (l0/(l0-l4)) \
                    * (exp((l0 - l4)*t) - 1.) + 1.
            du3 = (A[0] - 1.)*l0*exp(l0*t) - (A[0] - 1.) * (l0/(l0-l4)) \
                    * (l0-l4)*exp((l0-l4)*t)
        else:  # pres. [230Th/238U], pres. [234U/238U]
            u3 = 1. - (A[0] - 1.) * (l0/(l0-l4))*(exp(l0*t) - exp(l4*t)) \
                    + (A[1] - 1.)*exp(l0*t)
            du3 = l0 * (A[0] - 1.) * exp(l0 * t) - (A[0] - 1.) * (l0/(l0-l4)) \
                    * (l0 * exp(l0*t) - l4 * exp(l4*t))

    dF1 = l8 * exp(l8 * t) * (c1 * exp(-l8 * t) + c2 * exp(-l4 * t) + c3 * exp(-l0 * t)
                + c4*exp(-l6 * t) + 1) + exp(l8 * t) * (-c1 * l8 * exp(-l8 * t)
                - c2*l4*exp(-l4 * t) - c3*l0*exp(-l0 * t) - c4*l6*exp(-l6 * t))

    dF2 = du2 * v2 + u2 * dv2
    dF3 = du3 * v3 + u3 * dv3
    dF4 = l8 / l6 * A[2] * exp(l8 * t) * (l8 * (1 - exp(-l6 * t)) + (l6 * exp(-l6 * t)))

    if comp:
        return dF1, dF2, dF3, dF4
    return dF1 + dF2 + dF3 + dF4


def dgdt(t, A15i, comp=False):
    """
    !!! Experimental function !!!
    Derivative of g with respect to t.

    Parameters
    ----------
    comp : bool
        if True, then returns individual components

    """
    l5, l1 = cfg.lam235, cfg.lam231
    d1, d2 = bateman((l5, l1), series='235U')
    dG1 = l5 * exp(l5 * t) * (d1 * exp(-l5 * t) + d2 * exp(-l1 * t) + 1.) \
          + exp(l5 * t) * (-d1 * l5 * exp(-l5 * t) - l1 * d2 * exp(-l1 * t))
    dG2 = l5 / l1 * A15i * exp(l5 * t) * (l5 * (1 - exp(-l1 * t)) + (l1 * exp(-l1 * t)))
    if comp:
        return dG1, dG2
    return dG1 + dG2



"""
Minimisation functions for numerical disequilibrium age solutions.

References
----------
.. [Guillong2014]
    Guillong, M., von Quadt, A., Sakata, S., Peytcheva, I., Bachmann, O., 2014.
    LA-ICP-MS Pb-U dating of young zircons from the Kos–Nisyros volcanic centre,
    SE aegean arc. Journal of Analytical Atomic Spectrometry 29, 963–970.
    https://doi.org/10.1039/C4JA00009A
.. [Sakata2017]
    Sakata, S., Hirakawa, S., Iwano, H., Danhara, T., Guillong, M., Hirata, T.,
    2017. A new approach for constraining the magnitude of initial
    disequilibrium in Quaternary zircons by coupled uranium and thorium decay
    series dating. Quaternary Geochronology 38, 1–12.
    https://doi.org/10.1016/j.quageo.2016.11.002

"""

import numpy as np

from . import ludwig
from .. import misc


#=================================
# Disequilibrium Age minimisation functions
#=================================

def concage_x(diagram, t0=1.0, init=(False, False)):
    """ """
    assert diagram == 'tw'
    h = abs(t0) * np.sqrt(np.finfo(float).eps)
    def fmin(t, x, A, DC, BC):
        return x - 1. / ludwig.f(t, A, DC, BC, init=init)
    def dfmin(t, x, A, DC, BC):
        return misc.cdiff(t, fmin, h, x, A, DC, BC)
    return fmin, dfmin


def concint(t0=1.0, diagram='tw', init=(True, True)):
    """
    Disequilbrium concordia-intercept age minimisation function.
    """
    assert diagram in ('tw', 'wc')
    h = abs(t0) * np.sqrt(np.finfo(float).eps)
    if diagram == 'tw':
        def fmin(t, a, b, A238, A235, DC8, DC5, BC8, BC5, U):
            return b + a * ludwig.f(t, A238, DC8, BC8, init=init) \
                   - ludwig.g(t, A235, DC5, BC5) / U
        def dfmin(t, a, b, A238, A235, DC8, DC5, BC8, BC5, U):
            return misc.cdiff(t, fmin, h, a, b, A238, A235, DC8, DC5,
                    BC8, BC5, U)
    else:
        raise ValueError('not yet coded')
    return fmin, dfmin


def isochron(t0=1.0, age_type='iso-Pb6U8', init=(True, True)):
    """
    Disequilibrium isochron age minimisation function.
    """
    assert age_type in ('iso-Pb6U8', 'iso-Pb7U5')
    h = abs(t0) * np.sqrt(np.finfo(float).eps)
    if age_type == 'iso-Pb6U8':
        def fmin(t, b, A, DC, BC):
            return b - ludwig.f(t, A, DC, BC, init=init)
    elif age_type == 'iso-Pb7U5':
        def fmin(t, b, A, DC, BC):
            return b - ludwig.g(t, A, DC, BC)
    def dfmin(t, b, A, DC, BC):
        return misc.cdiff(t, fmin, h, b, A, DC, BC)
    return fmin, dfmin


def concordant_A48():
    """
    Minimisation function for computing initial U234/U238 activity ratio that
    forces concordance between 238U and 235U isochron ages.

    Minimises function: f(t75, A) - slope_86, where t75 is
    the 207Pb/x-235U/x isochron age.
    """
    h = 1.
    def fmin(A48i, t57, slope_86, A, DC, BC):
        return ludwig.f(t57, [A48i, A[1], A[2]], DC, BC) - slope_86
    def dfmin(A48i, t57, slope_86, A, DC, BC):
        return misc.cdiff(A48i, fmin, h, t57, slope_86, A, DC, BC)
        # return DC[0] / DC[1] * np.exp(DC[0] * t57) * (BC[3] * np.exp(-DC[1] * t57)
        #        + (BC[4] * np.exp(-DC[1] * t57) + 1.))
    return fmin, dfmin


def pbu(age_type='Pb6U8', t0=1., init=(True, True)):
    """
    Disequilbrium Pb/U age functions.
    """
    h = abs(t0) * np.sqrt(np.finfo(float).eps)
    if age_type == 'Pb6U8':
        # x is 206Pb/238U or 207Pb/235U
        def fmin(t, x, A, DC, BC):
            return x - ludwig.f(t, A, DC, BC, init=init)
        def dfmin(t, x, A, DC, BC):
            return misc.cdiff(t, fmin, h, x, A, DC, BC)
    elif age_type == 'Pb7U5':
        def fmin(t, x, A, DC, BC):
            return x - ludwig.g(t, A, DC, BC)
        def dfmin(t, x, A, DC, BC):
            return misc.cdiff(t, fmin, h, x, A, DC, BC)
    return fmin, dfmin


def guillong(t0=1.):
    """
    Minimisation function for disequilibrium Pb/U ages using the equations of
    Guillong et al., (2014) - eith 206Pb/238U or 207Pb/235U ages.
    fXU is either fThU or fPaU
    lamU is either lam238 or lam235
    lamX is either lam230 or lam231
    """
    def fmin(t, x, fXU, lamU, lamX):
        """ Age minisation functions using equations of Guillong (2014).
        """
        return (np.exp(lamU * t) - 1.) + lamU / lamX \
               * (fXU - 1.) * (1. - np.exp(-lamX * t)) \
               * np.exp(lamU * t) - x
    def dfmin(t, x, fXU, lamU, lamX):
        """x is dummy argument
        """
        return lamU * np.exp(lamU * t) + lamU / lamX * (fXU - 1.) \
               * (lamU - (lamU - lamX) * np.exp(- lamX * t)) \
               * np.exp(lamU * t)
    return fmin, dfmin


def mod207(t0=1., init=(True, True)):
    """
    Age minimisation function for modified 207Pb ages.
    """
    h = abs(t0) * np.sqrt(np.finfo(float).eps)
    def fmin(t, x, y, A, DC8, DC5, BC8, BC5, U, Pb76):
        """
        x : measured 206Pb*/238U
        y : measured 207Pb*/206Pb
        """
        num = ludwig.g(t, A[-1], DC5, BC5) / U - (y - Pb76) / x
        denom = ludwig.f(t, A[:-1], DC8, BC8, init=init)
        return num / denom - Pb76
    def dfmin(t, x, y, A, DC8, DC5, BC8, BC5, U, Pb76):
        out = misc.cdiff(t, fmin, h, x, y, A, DC8, DC5, BC8, BC5, U, Pb76)
        return out
    return fmin, dfmin


def sakata(t0=1.):
    """
    Pb/U age minimisation functions using equaitons of Sakata (2017).
    """
    h = abs(t0) * np.sqrt(np.finfo(float).eps)
    def fmin(t, x, y, fThU, fPaU, lam238, lam230, lam235, lam231, U, Pb76):
        F = np.exp(lam238 * t) - 1. \
              + lam238 / lam230 * (fThU - 1) * (1 - np.exp(-lam230 * t)) \
              * np.exp(lam238 * t)
        G = (np.exp(lam235 * t) - 1.) \
              + lam235 / lam231 * (fPaU - 1) * (1 - np.exp(-lam231 * t)) \
              * np.exp(lam235 * t)
        num = (1 / U) * G - (y - Pb76) / x
        return num / F - Pb76
    def dfmin(t, x, y, fThU, fPaU, lam238, lam230, lam235, lam231, U, Pb76):
        return misc.cdiff(t, fmin, h, x, y, fThU, fPaU, lam238, lam230, lam235,
                          lam231, U, Pb76)
    return fmin, dfmin




from setuptools import setup

with open("README.md", "r") as fh:
   long_description = fh.read()

setup(
	name='pysoplot',
	version='0.0.1',
	description='a Python geochronology library',
	license='MIT',
	packages=['pysoplot', 'pysoplot.uxpb'],
	package_dir={'': 'src'},
	url="https://github.com/timpol/pysoplot",
	author="Timothy Pollard",
	author_email="pollard@student.unimelb.edu.au",
	platforms="all",
	classifiers=[
		"Development Status :: 4 - Beta",
		"Intended Audience :: Science/Research",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
		"Operating System :: MacOS :: MacOS X",
		"Operating System :: Microsoft :: Windows",
		"Operating System :: POSIX :: Linux",
		"Programming Language :: Python :: 3",
		"Programming Language :: Python :: 3 :: Only",
		"Programming Language :: Python :: 3.7",
		"Programming Language :: Python :: 3.8",
		"Programming Language :: Python :: 3.9",
		"Programming Language :: Python :: 3.10",
	],
	long_description=long_description,
	long_description_content_type="text/markdown",
	install_requires=[
		"numpy",
		"scipy",
		"matplotlib"
	],
	extras_require={
		"dev": []
	}
)



"""
Functions for calculating and plotting U-Pb ages accounting for disequilibrium
amongst intermediate nuclides of the uranium decay series.

"""

from . import dqmc
from . import dqpb
from . import guillong
from . import ludwig
from . import minimise
from . import wendt_carl




"""
Plotting routines and functions.

Notes
------
Currently based on matplotlib. In future core functions may be re-written
to make them plotting libray independent.

References
----------
.. [Ludwig1980]
    Ludwig, K.R., 1980. Calculation of uncertainties of U-Pb isotope data. Earth
    and Planetary Science Letters 212–202.
.. [Noda2017]
    Noda, A., 2017. A new tool for calculation and visualization of U. Bulletin of
    the Geological Survey of Japan 1–10.

"""


import warnings
import numpy as np
import matplotlib.pyplot as plt

from matplotlib import ticker
from matplotlib import transforms
from matplotlib.patches import Ellipse
from matplotlib.colors import to_rgba
from scipy import stats

from . import cfg
from . import misc


#=================================
# 2-d data plotting
#=================================

def rline(ax, theta):
    """
    Add regression line to plot.
    """
    xmin, xmax = ax.get_xlim()
    ax.plot([xmin, xmax],
            [np.dot(theta, [1, xmin]), np.dot(theta, [1, xmax])],
            label='regression line', **cfg.rline_kw)


def renv(ax, fit):
    """
    Plot error envelope about regression line showing 95% confidence
    limits on fit. For classical regression fits, uses the approach of
    Ludwig (1980), for non-classical fits uses Monte Carlo simulation.
    """
    # TODO: this approach fails when x at ymin and x at ymax are very close (?)

    if fit['type'] == 'classical':
        xx = np.linspace(*ax.get_xlim(), num=100, endpoint=True)

        a, b = fit['theta']
        a_95pm, b_95pm = fit['theta_95ci']
        xbar = fit['xbar']

        # Eq. 29 of Ludwig (1980); EPSL.
        ym = a + b * xx
        yhigh = ym + np.sqrt(a_95pm ** 2 + b_95pm ** 2 * xx * (xx - 2. * xbar))
        ylow = ym - np.sqrt(a_95pm ** 2 + b_95pm ** 2 * xx * (xx - 2. * xbar))

        ax.plot(xx, ylow, label='regression envelope line', **cfg.renv_line_kw)
        ax.plot(xx, yhigh, label='regression envelope line', **cfg.renv_line_kw)
        ax.fill_between(xx, yhigh, ylow, label='regression envelope', **cfg.renv_kw)

    else:
        # use Monte Carlo approach:
        renv_mc(ax, fit)


def renv_mc(ax, fit):
    """
    Use Monte Carlo simulation to plot regression error envelope.
    """
    npts = 100                  # number of points to construct curve
    ns = 10_000                 # number of simulated theta

    theta = fit['theta']
    covtheta = fit['covtheta']

    # simulate slope and intercept values
    ths = cfg.rng.multivariate_normal(theta, covtheta, ns)

    # Find intercept points b/w regression line and axis frame, then
    # sort these by increasing y.
    w_intercepts = box_line_intersection(*ax.get_xlim(), *ax.get_ylim(), *theta)
    w_intercepts = w_intercepts[w_intercepts[:, 1].argsort()]

    if w_intercepts.size == 0:
        warnings.warn('regression line envelope could not be plotted as it appears to lie outside axis limits')
        return

    # If line intercepts window at ymin and ymax, then simulate x for given
    # vector of evenly spaced y spanning axis window:
    if (np.isclose(w_intercepts[0, 1], ax.get_ylim()[0]) and
            np.isclose(w_intercepts[1, 1], ax.get_ylim()[1])):

        ymin = w_intercepts[0, 1]
        ymax = w_intercepts[1, 1]
        yspan = ymax - ymin
        # increase span a bit to avoid edge effects
        yq = np.linspace(ymin - 1.25 * yspan, ymax + 1.25 * yspan, num=npts,
                         endpoint=True)

        # calculate simulated x for each y and theta combination
        # x = (y - a) / b over each y and a, b combination
        xq = np.tile(yq, (ns, 1)).T - ths[:, 0]
        xq /= np.tile(ths[:, 1], (npts, 1))

        # get upper and lower limits
        xhigh = np.quantile(xq, 0.975, axis=1)
        xlow = np.quantile(xq, 0.025, axis=1)

        ax.plot(xlow, yq, label='regression envelope line', **cfg.renv_line_kw)
        ax.plot(xhigh, yq, label='regression envelope line', **cfg.renv_line_kw)
        ax.fill_betweenx(yq, xlow, xhigh, label='regression envelope', **cfg.renv_kw)

    # Simulate y for vector of evenly spaced x:
    else:
        xmin = w_intercepts[0, 0]
        xmax = w_intercepts[1, 0]
        xspan = xmax - xmin
        # increase span a bit to avoid edge effects
        xq = np.linspace(xmin - 1.25 * xspan, xmax + 1.25 * xspan, num=npts,
                         endpoint=True)

        # calculate simulated y values
        # y = theta . X
        yq = ths @ np.array((np.ones(npts), xq))

        # get upper and lower limits
        yhigh = np.quantile(yq, 0.975, axis=0)
        ylow = np.quantile(yq, 0.025, axis=0)

        ax.plot(xq, ylow, label='regression envelope line', **cfg.renv_line_kw)
        ax.plot(xq, yhigh, label='regression envelope line', **cfg.renv_line_kw)
        ax.fill_between(xq, yhigh, ylow, label='regression envelope', **cfg.renv_kw)


def plot_rfit(ax, fit):
    """Add regression line and regression error envelope to plot in one go.
    """
    rline(ax, fit['theta'])
    renv(ax, fit)


def plot_mod207_projection(ax, x, sx, y, sy, r_xy, Pb76):
    """  """
    n = x.shape[0]
    for i in range(n):
        a = Pb76
        b = (y[i] - a) / x[i]
        theta = (a, b)
        ax.plot([0, x[i]],
                [np.dot(theta, [1., 0.]), np.dot(theta, [1., x[i]])],
                label='projection line', **cfg.pb76_line_kw)


def plot_dp(x, sx, y, sy, r_xy, labels=None, p=0.95, reset_axis_limits=True):
    """
    Plot 2-dimensional data points as confidence ellipses.

    Notes
    ------
    Assumes a large sample size was used to estimate data point uncertainties
    and covariances.

    """
    if labels is not None:
        assert len(labels) == x.shape[0]

    fig, ax = plt.subplots(**cfg.fig_kw, subplot_kw=cfg.subplot_kw)
    n = x.shape[0]

    for i in range(n):
        confidence_ellipse(ax, x[i], sx[i], y[i], sy[i], r_xy[i], p=p,
                           ellipse_kw=cfg.dp_ellipse_kw,
                           mpl_label='data ellipse')
        if labels is not None:
            ax.annotate(labels[i], (x[i], y[i]), **cfg.dp_label_kw,
                        label='data ellipse label')

    # set axis limits
    if reset_axis_limits:
        fig.canvas.draw()
        inv = ax.transData.inverted().transform        # display to data coords
        xy = np.array([inv(e.get_extents().get_points()).flatten() for e in
                       ax.patches if e.get_label() == 'data ellipse'])
        xmin = np.min(xy[:, 0::2])
        xmax = np.max(xy[:, 0::2])
        ymin = np.min(xy[:, 1::2])
        ymax = np.max(xy[:, 1::2])
        xpad = (xmax - xmin) / 5
        ypad = (ymax - ymin) / 5
        ax.set_xlim((xmin - xpad, xmax + xpad))
        ax.set_ylim((ymin - ypad, ymax + ypad))

        # turn off to prevent further auto updates (is this needed?)
        ax.autoscale(enable=False, axis='both')

    return fig


def confidence_ellipse(ax, x, sx, y, sy, r_xy, p=0.95, mpl_label='data ellipse',
                       ellipse_kw=None, outline_alpha=False):
    """
    Plot 2D correlated data point as covariance ellipse. Assumes large
    sample size was used to estimate data point uncertainties and covariance.

    Based partly on code implemented in Knighton and Tobin (2011), see:
    https://github.com/joferkington/oost_paper_code/blob/master/error_ellipse.py

    """
    if ellipse_kw is None:
        ellipse_kw = {}

    df = 2
    s = stats.chi2.ppf(p, df)

    # Compile covariance matrix
    cov = r_xy * sx * sy
    cm = np.array([[sx ** 2, cov], [cov, sy ** 2]])

    if not misc.pos_def(cm):
        raise ValueError('data point covariance matrix is not positive definite')

    # Get sorted eigenvalues and eigenvectors of variance-covariance matrix.
    w, v = misc.eigsorted(cm)

    # Calculate major and minor axis lengths.
    lx, ly = 2 * np.sqrt(s * w)

    # Angle of principle eigenvector from positive x-axis.
    theta = np.degrees(np.arctan2(*v[:, 0][::-1]))

    # Ignore alpha when plotting ellipse outline
    if outline_alpha == False:
        ec = None
        fc = None
        for k, v in ellipse_kw.items():
            a = 1. if (not 'alpha' in ellipse_kw.keys()) else ellipse_kw['alpha']
            if k in ('ec', 'edgecolor'):
                v = to_rgba(v)[:3]
                ec = to_rgba(v, 1.0)
            elif k in ('fc', 'facecolor'):
                v = to_rgba(v)[:3]
                fc = to_rgba(v, a)
        if ec is not None:
            ellipse_kw['ec'] = ec
        if fc is not None:
            ellipse_kw['fc'] = fc

    ellipse = Ellipse((x, y), width=lx, height=ly, angle=theta,
                      label=mpl_label, **ellipse_kw)
    return ax.add_patch(ellipse)


def confidence_ellipse2(ax, x, y, cov, n_std=2.0, **ellipse_kw):
    """
    Alternative confidence ellipse routine from:
    https://matplotlib.org/devdocs/gallery/statistics/confidence_ellipse.html

    Create a plot of the covariance confidence ellipse of *x* and *y*.
    n_std : The number of standard deviations to determine the ellipse's
            radii.

    CAREFUL : is the scaling here correct for a true confidence ellipse? definition
    from the doc string seems incorrect.

    """
    if x.size != y.size:
        raise ValueError("x and y must be the same size")

    if not misc.pos_def(cov):
        raise ValueError('data point covariance matrix is not positive definite')

    # cov = np.cov(x, y)
    pearson = cov[0, 1] / np.sqrt(cov[0, 0] * cov[1, 1])
    # Using a special case to obtain the eigenvalues of this
    # two-dimensionl dataset.
    ell_radius_x = np.sqrt(1 + pearson)
    ell_radius_y = np.sqrt(1 - pearson)
    ellipse = Ellipse((0, 0), width=ell_radius_x * 2, height=ell_radius_y * 2,
                      **ellipse_kw)

    # Calculating the stdandard deviation of x from
    # the squareroot of the variance and multiplying
    # with the given number of standard deviations.
    scale_x = np.sqrt(cov[0, 0]) * n_std
    mean_x = np.mean(x)

    # calculating the stdandard deviation of y ...
    scale_y = np.sqrt(cov[1, 1]) * n_std
    mean_y = np.mean(y)

    transf = transforms.Affine2D() \
        .rotate_deg(45) \
        .scale(scale_x, scale_y) \
        .translate(mean_x, mean_y)

    ellipse.set_transform(transf + ax.transData)
    return ax.add_patch(ellipse)


#=================================
# Weighted average plots
#=================================

def wav_plot(x, xpm, xb, xbpm, rand_pm=None, sorted=False, ylim=(None, None),
             x_multiplier=1., dp_labels=None):
    """
    Plot 1d data points and weighted average as line and uncertainty band.

    Parameters
    ----------
    x : np.ndarray
        values to be averaged
    xpm : np.ndarray
        symmetrical +/- errors on x
    xb : float
        average value (x-bar)
    xbpm : float
        symmetrical +/- error on xb
    rand_pm : np.ndarray, optional
        random symmetrical +/- errors on x
    x_multiplier :
        Use for unit conversion (e.g. set to 1000 for Ma to ka)
    dp_labels : array-like, optional
        data point labels

    """
    assert all([x >= 0 for x in xpm])
    fig, ax = plt.subplots(**cfg.wav_fig_kw, subplot_kw=cfg.subplot_kw)
    if x_multiplier != 1.:
        x = x * x_multiplier
        xpm = xpm * x_multiplier
        xb = xb * x_multiplier
        xbpm = xbpm * x_multiplier
        if rand_pm is not None:
            rand_pm = rand_pm * x_multiplier
    oned_dp(ax, x, xpm, rand_pm=rand_pm, sorted=sorted, labels=dp_labels)
    wav_line(ax, xb, xbpm)
    apply_plot_settings(fig, plot_type='wav', ylim=ylim)
    return fig


def oned_dp(ax, x, xpm, rand_pm=None, sorted=False, labels=None):
    """
    Plot 1-d data points (e.g. ages used to compute weighted mean).
    """
    assert len(x) == len(xpm)
    n = x.shape[0]
    if rand_pm is not None:
        assert len(x) == len(rand_pm)
    if sorted:
        ind = np.argsort(x)
        x = x[ind]
        xpm = xpm[ind]
        if rand_pm is not None:
            rand_pm = rand_pm[ind]

    ind = np.arange(n)

    # set x-limits
    xpad = cfg.wav_marker_width / 2
    ax.set_xlim((-1 + xpad, n - xpad))

    if rand_pm is not None:
        # random errors
        heights = 2. * rand_pm
        bottoms = x - rand_pm
        ax.bar(ind, heights, bottom=bottoms, width=cfg.wav_marker_width,
               **cfg.wav_markers_rand_kw, label='wav marker rand')
        # systematic errors
        if not all(xpm > rand_pm):
            warnings.warn('random only analytical errors were greater than '
                          'totol errors for one or more data points')
        # upper segment:
        heights = xpm - rand_pm
        ax.bar(ind, heights, bottom=(x + rand_pm), width=cfg.wav_marker_width,
               **cfg.wav_markers_kw, label='wav marker full')
        # lower segment:
        ax.bar(ind, heights, bottom=(x - xpm), width=cfg.wav_marker_width,
               **cfg.wav_markers_kw, label='wav marker full')
    else:
        # full errors
        heights = 2. * xpm
        bottoms = x - xpm
        ax.bar(ind, heights, bottom=bottoms, width=cfg.wav_marker_width,
               **cfg.wav_markers_rand_kw, label='wav marker full')

    # set y-axis limits
    top = np.max(x + xpm)
    bottom = np.min(x - xpm)
    yspread = top - bottom
    ypad = 0.1 * yspread
    ax.set_ylim((bottom - ypad, top + ypad))

    # add labels
    if labels is not None:
        ax.set_xticks(ind)
        # Set ticks labels for x-axis
        ax.set_xticklabels(labels, rotation=45)
    else:
        # remove x-ticks/labels
        ax.xaxis.set_major_locator(plt.NullLocator())


def wav_line(ax, xb, xbpm, env=True):
    """ Add weighted average line and error envelope to plot of 1-D data points.
    """
    # plot weighted mean
    ax.axhline(y=xb, **cfg.wav_line_kw, label='wave line')
    # plot error envelope about weighted mean
    if env:
        y_low = xb - xbpm
        y_hi = xb + xbpm
        ax.axhspan(y_low, y_hi, **cfg.wav_env_kw, label='wav env')


#=================================
# Plot settings and formatting
#=================================

def apply_plot_settings(fig, plot_type='isochron', diagram=None,
                xlim=(None, None), ylim=(None, None), axis_labels=(None, None),
                tight_layout=True, norm_isotope='204Pb'):
    """
    Set labels, axis limits, and apply label, tick, grid formatting settings
    from cfg.
    """
    assert plot_type in ('isochron', 'intercept', 'wav', 'hist')

    # ax = fig.axes[0]
    for ax in fig.axes:
        # set axis limits
        set_axis_limits(ax, xmin=xlim[0], xmax=xlim[1], ymin=ylim[0], ymax=ylim[1])

        # set axis labels
        if diagram is not None:
            set_axis_labels(ax, diagram=diagram, norm_isotope=norm_isotope,
                        axis_labels=axis_labels)

        # # set spine params
        # for pos in ['left', 'right', 'bottom', 'top']:
        #     ax.spines['left'].set(**cfg.spine_kw)

        # Set spine options
        if len(cfg.spine_kw.keys()) > 0:
            plt.setp(ax.spines.values(), **cfg.spine_kw)

        # Set major tick properties.
        if plot_type == 'wav':
            ax.tick_params(axis='y', which='major', **cfg.major_ticks_kw)
            # Set minor tick properties.
            if cfg.show_minor_ticks:
                ax.minorticks_on()
                ax.tick_params(axis='y', which='minor', **cfg.minor_ticks_kw)
        else:
            ax.tick_params(axis='both', which='major', **cfg.major_ticks_kw)
            # Set minor tick properties.
            if cfg.show_minor_ticks:
                ax.minorticks_on()
                ax.tick_params(axis='both', which='minor', **cfg.minor_ticks_kw)

        # Set tick label formats
        if len(ax.get_xticks()) > 0 and plot_type != 'wav':
            xtick_formatter = tick_label_format(ax, ax.get_xticks(),
                    cfg.sci_limits, cfg.comma_sep_thousands, axis='x')
            ax.get_xaxis().set_major_formatter(xtick_formatter)

        if len(ax.get_yticks()) > 0:
            ytick_formatter = tick_label_format(ax, ax.get_yticks(),
                    cfg.sci_limits, cfg.comma_sep_thousands, axis='y')
            ax.get_yaxis().set_major_formatter(ytick_formatter)

        #  Major grid lines
        if cfg.show_major_gridlines:
            if plot_type == 'wav':
                ax.grid(True, which='major', axis='y', **cfg.gridlines_kw)
            else:
                ax.grid(True, which='major', axis='both', **cfg.gridlines_kw)
            ax.set_axisbelow(True)

        # Minor grid lines
        if cfg.show_minor_gridlines:
            if plot_type == 'wav':
                ax.grid(True, which='minor', axis='y', **cfg.gridlines_kw)
            else:
                ax.grid(True, which='minor', axis='both', **cfg.gridlines_kw)
            ax.set_axisbelow(True)

        # Hide spines
        if cfg.hide_top_spine:
            ax.spines['right'].set_visible(False)
            # hide ticks as well
            ax.tick_params(top=False)

        if cfg.hide_right_spine:
            ax.spines['top'].set_visible(False)
            # hide ticks as well
            ax.tick_params(right=False)

        # Set axis and tick label font sizes
        # note: this axis labels now done by set_axis_labels() function
        ax.xaxis.set_tick_params(labelsize=cfg.tick_label_size)
        ax.yaxis.set_tick_params(labelsize=cfg.tick_label_size)

        # Set axis exponent multiplier fontsize
        t = ax.yaxis.get_offset_text()
        t.set_size(cfg.exp_font_size)
        t = ax.xaxis.get_offset_text()
        t.set_size(cfg.exp_font_size)

    # tight layout
    fig.tight_layout()


def set_axis_limits(ax, xmin=None, xmax=None, ymin=None, ymax=None):
    """
    Set axis limits if user specified limits are supplied.
    """
    xmin0, xmax0 = ax.get_xlim()
    ymin0, ymax0 = ax.get_ylim()

    # Set axis limits to user defined values.
    if xmin is not None:
        if xmax is not None:
            if xmin >= xmax:
                raise ValueError('xmin cannot be greater than xmax')
        elif xmin > xmax0:
            raise ValueError('xmin cannot be greater than xmax')
        ax.set_xlim(left=xmin)

    if xmax is not None:
        if xmin is not None:
            if xmin >= xmax:
                raise ValueError('xmin cannot be greater than xmax')
        elif xmax < xmin0:
            raise ValueError('xmin cannot be greater than xmax')
        ax.set_xlim(right=xmax)

    if ymin is not None:
        if ymax is not None:
            if ymin >= ymax:
                raise ValueError('ymin cannot be greater than ymax')
        elif ymin > ymax0:
            raise ValueError('ymin cannot be greater than ymax')
        ax.set_ylim(bottom=ymin)

    if ymax is not None:
        if ymin is not None:
            if ymin >= ymax:
                raise ValueError('ymin cannot be greater than ymax')
        elif ymax < ymin0:
            raise ValueError('ymin cannot be greater than ymax')
        ax.set_ylim(top=ymax)


def set_axis_labels(ax, diagram='tw', norm_isotope='204Pb',
                    axis_labels=(None, None)):
    """
    Set axis x, y axis labels for default diagram types, or use inputted
    axis labels.
    """
    xlabel, ylabel = None, None

    if diagram == 'Pb6U8':
        ylabel = "$^{206}$Pb/$^{238}$U age (Ma)"

    elif diagram == 'Pb7U5':
        ylabel = "$^{207}$Pb/$^{235}$U age (Ma)"

    elif diagram == 'mod-207Pb':
        ylabel = "Modified $^{207}$Pb age (Ma)"

    elif diagram == "tw":
        xlabel = "$^{238}$U/$^{206}$Pb"
        ylabel = "$^{207}$Pb/$^{206}$Pb"

    elif diagram == 'wc':
        xlabel = "$^{207}$Pb/$^{235}$U"
        ylabel = "$^{206}$Pb/$^{238}$U"

    elif diagram == 'iso-Pb6U8' and norm_isotope == '208Pb':
        xlabel = "$^{238}$U/$^{208}$Pb"
        ylabel = "$^{206}$Pb/$^{208}$Pb"

    elif diagram == 'iso-Pb6U8' and norm_isotope == '204Pb':
        xlabel = "$^{238}$U/$^{204}$Pb"
        ylabel = "$^{206}$Pb/$^{204}$Pb"

    elif diagram == 'iso-Pb7U5' and norm_isotope == '208Pb':
        xlabel = "$^{235}$U/$^{208}$Pb"
        ylabel = "$^{207}$Pb/$^{208}$Pb"

    elif diagram == 'iso-Pb7U5' and norm_isotope == '204Pb':
        xlabel = "$^{235}$U/$^{204}$Pb"
        ylabel = "$^{207}$Pb/$^{204}$Pb"

    if axis_labels[0] is not None:
        xlabel = axis_labels[0]
    if axis_labels[1] is not None:
        ylabel = axis_labels[1]

    if xlabel is not None:
        ax.set_xlabel(xlabel, **cfg.axis_labels_kw)
    if ylabel is not None:
        ax.set_ylabel(ylabel, **cfg.axis_labels_kw)


def tick_label_format(ax, ticks, sci_limits, comma_sep_thousands, axis='x'):
    """
    Format plot tick labels.

    Parameters
    -------------
    ticks : array-like
        List of ticks.
    sci_limits : array-like
        Axis scientific notation exponent limits as [lower, upper]. If the
        maximum axis value is greater than 10 ^ upper then scientific notation
        is used for tick labels. An equivalent test is applied to the lower
        limit.
    comma_sep_thousands : bool
        True if use comma to separate each 10^3 increment, e.g. "10,450".

    Returns
    -------
    tick_formatter :  matplotlib.ticker.Formatter
    """
    # Use scientiffic notation if axis values outside power limits.
    if misc.get_exponent(min(ticks)) < sci_limits[0] or \
            misc.get_exponent(max(ticks)) > sci_limits[1]:
        # TODO: consider improving this following e.g.:
        # <https://stackoverflow.com/questions/31517156/adjust-exponent-text-
        # after-setting-scientific-limits-on-matplotlib-axis>
        tick_formatter = ticker.ScalarFormatter(useMathText=True)
        tick_formatter.set_powerlimits((0, 0))

        if axis == 'x':
            ax.xaxis.get_offset_text().set_fontsize(
                ax.get_xticklabels()[0].get_fontsize() + 1)
        else:
            ax.yaxis.get_offset_text().set_fontsize(
                ax.get_yticklabels()[0].get_fontsize() + 1)

        return tick_formatter

    elif comma_sep_thousands:
        # Find number of decimal digits to show.
        ndp = 0
        for z in ticks:
            # mpl should give nice values, but just in case of rounding issues.
            z = misc.round_down(float(z), 12)
            nj = misc.num_dec_places(z)
            if nj > ndp:
                ndp = nj
        # format with comma sep thousands
        tick_formatter = ticker.StrMethodFormatter('{{x:,.{}f}}'.format(ndp))
        return tick_formatter

    else:
        # Just use standard formatting without sci notation.
        tick_formatter = ticker.ScalarFormatter()
        tick_formatter.set_powerlimits((-np.inf, np.inf))
        return tick_formatter


#=======================================
# Plot intersection and bbox functions
#=======================================

def box_line_intersection(xmin, xmax, ymin, ymax, a, b):
    """
    Find intercpet points between line defined by slope = a, int = b, and
    axis window limits.
    """
    p = np.full((4, 2), np.nan)
    bbox_lines = [[(xmin, ymax), (xmax, ymax)],
                  [(xmin, ymin), (xmax, ymin)],
                  [(xmin, ymin), (xmin, ymax)],
                  [(xmax, ymin), (xmax, ymax)]]
    for i, r in enumerate(bbox_lines):
        p[i] = line_intersection(a, b, r)
    p = p[np.any(~np.isnan(p), axis=1)]
    # can return repeated intercept points if they are at the box corners !!
    if p.shape[0] > 2:
        p = p[:2]
    return p


def line_intersection(a1, b1, L2):
    """
    Return intersection point between line defined by slope, b1, and y-int,
    a1, and a line segment defined by end points as L2 = ((x1, y1), (x2, y2)).
    """
    assert np.array(L2).shape == (2, 2)
    if L2[1][0] - L2[0][0] == 0:  # if b2 = inf.
        p = np.empty(2)
        p[0] = L2[1][0]
        p[1] = a1 + b1 * p[0]

    else:
        b2 = (L2[1][1] - L2[0][1]) / (L2[1][0] - L2[0][0])
        a2 = L2[0][1] - b2 * L2[0][0]

        A = np.array([[-a1],
                      [-a2]])
        B = np.array([[b1, -1.],
                      [b2, -1.]])

        # calculate intersection point
        p = np.dot(np.linalg.inv(B), A)
        p = p.flatten()

    # check if intersection within end points, otherwise return nan
    xmin = min(L2[0][0], L2[1][0])
    xmax = max(L2[0][0], L2[1][0])
    ymin = min(L2[0][1], L2[1][1])
    ymax = max(L2[0][1], L2[1][1])

    p = p if in_box(p, xmin, xmax, ymin, ymax) else np.nan
    return p


def in_box(p, xmin, xmax, ymin, ymax):
    """
    Check if point p = (x, y) is within bounding rectangle
    defined by x, y limits, allowing some tolerance for floating points numbers.
    """
    assert not xmin > xmax and not ymin > ymax
    if ((p[0] > xmin or np.isclose(p[0], xmin)) and
        (p[0] < xmax or np.isclose(p[0], xmax)) and
        (p[1] > ymin or np.isclose(p[1], ymin)) and
        (p[1] < ymax or np.isclose(p[1], ymax))):
        return True

    return False
